from loguru import logger

from veros.diagnostics.diagnostic import VerosDiagnostic
from veros import veros_method


class NPZDMonitor(VerosDiagnostic):
    """Diagnostic monitoring nutrients and plankton concentrations
    """

    name = 'npzd'  #:
    output_frequency = None  #: Frequency (in seconds) in which output is written
    restart_attributes = []
    save_graph = False  #: Whether or not to save a graph of the selected dynamics
    graph_attr = {  #: Properties of the graph (graphviz)
        'splines': 'ortho',
        'center': 'true',
        'nodesep': '0.05',
        'node': 'square'
    }

    def __init__(self, setup):
        self.output_variables = []
        self.surface_out = []
        self.bottom_out = []
        self.po4_total = 0
        self.dic_total = 0

    @veros_method
    def initialize(self, vs):
        cell_volume = vs.area_t[2:-2, 2:-2, np.newaxis] * vs.dzt[np.newaxis, np.newaxis, :] * vs.maskT[2:-2, 2:-2, :]

        po4_sum = vs.phytoplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_PN\
                  + vs.detritus[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_PN\
                  + vs.zooplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_PN\
                  + vs.po4[2:-2, 2:-2, :, vs.tau]

        self.po4_total = np.sum(po4_sum * cell_volume)

        if vs.enable_carbon:
            dic_sum = vs.phytoplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_CN\
                      + vs.detritus[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_CN\
                      + vs.zooplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_CN\
                      + vs.dic[2:-2, 2:-2, :, vs.tau]

            self.dic_total = np.sum(dic_sum * cell_volume)

    def diagnose(self, vs):
        pass

    @veros_method
    def output(self, vs):
        """Print NPZD interaction graph
        """
        if self.save_graph:
            from graphviz import Digraph
            npzd_graph = Digraph('npzd_dynamics', filename='npzd_dynamics.gv')
            label_prefix = '\\tiny '  # should be selectable in settings allows for better exporting to tex
            label_prefix = ''

            # Create a node for all selected tracers
            # Drawing edges also creates nodes, so this just ensures, we se it,
            # when there are no connections to a node
            for tracer, tracer_data in vs.npzd_tracers.items():
                npzd_graph.node(tracer)

                # If a tracer has the sinking_speed attribute indicate on the graph
                if hasattr(tracer_data, 'sinking_speed'):
                    npzd_graph.node('Bottom', shape='square')
                    npzd_graph.edge(tracer, 'Bottom', label=label_prefix + 'sinking', lblstyle='sloped,above')

            # Common source rules are split up into several rules
            # This causes a duplication of the first registerd rule
            # Which should not be shown, so if there is more than one of them selected,
            # The source edge should not be displayed, because it is already there
            skiprules = []
            for name, rule in vs.common_source_rules.items():
                # Construct the list of rule names
                rule_names = [name + '_' + rule[0][1]] + [name + '_' + rule[i][2] for i in range(1, len(rule))]

                # If there is more than one indicate, that it should be skipped from drawing
                if len(rule_names) > 1:
                    skiprules.append(vs.npzd_available_rules[rule_names[0]])

            # Draw primary rules
            for rule in vs.npzd_rules:
                if rule in skiprules:
                    continue

                npzd_graph.edge(rule.source, rule.sink, label=label_prefix + rule.label, lblstyle='sloped, above')

            # Draw pre rules dotted
            for rule in vs.npzd_pre_rules:
                if rule in skiprules:
                    continue

                npzd_graph.edge(rule.source, rule.sink, label=label_prefix + rule.label, style='dotted', lblstyle='sloped, above')

            # Draw post rules dashed
            for rule in vs.npzd_post_rules:
                if rule in skiprules:
                    continue

                npzd_graph.edge(rule.source, rule.sink, label=label_prefix + rule.label, style='dashed', lblstyle='sloped, above')

            self.save_graph = False
            npzd_graph.render('npzd_graph', view=False)

        """
        Total phosphorus should be (approximately) constant
        """
        cell_volume = vs.area_t[2:-2, 2:-2, np.newaxis] * vs.dzt[np.newaxis, np.newaxis, :] * vs.maskT[2:-2, 2:-2, :]

        po4_sum = vs.phytoplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_PN\
                  + vs.detritus[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_PN\
                  + vs.zooplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_PN\
                  + vs.po4[2:-2, 2:-2, :, vs.tau]

        if vs.enable_carbon:
            dic_sum = vs.phytoplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_CN\
                      + vs.detritus[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_CN\
                      + vs.zooplankton[2:-2, 2:-2, :, vs.tau] * vs.redfield_ratio_CN\
                      + vs.dic[2:-2, 2:-2, :, vs.tau]

        po4_total = np.sum(po4_sum * cell_volume)
        logger.diagnostic(' total phosphorus: {}, relative change: {}'.format(po4_total, (po4_total - self.po4_total)/self.po4_total))
        self.po4_total = po4_total[...]

        if vs.enable_carbon:
            dic_total = np.sum(dic_sum * cell_volume)
            logger.diagnostic(' total DIC: {}, relative change: {}'.format(dic_total, (dic_total - self.dic_total)/self.dic_total))
            self.dic_total = dic_total.copy()

        for var in self.output_variables:
            if var in vs.recycled:
                recycled_total = np.sum(vs.recycled[var][2:-2, 2:-2, :] * cell_volume)
            else:
                recycled_total = 0

            if var in vs.mortality:
                mortality_total = np.sum(vs.mortality[var][2:-2, 2:-2, :] * cell_volume)
            else:
                mortality_total = 0

            if var in vs.net_primary_production:
                npp_total = np.sum(vs.net_primary_production[var][2:-2, 2:-2, :] * cell_volume)
            else:
                npp_total = 0

            if var in vs.grazing:
                grazing_total = np.sum(vs.grazing[var][2:-2, 2:-2, :] * cell_volume)
            else:
                grazing_total = 0

            logger.diagnostic(' total recycled {}: {}'.format(var, recycled_total))
            logger.diagnostic(' total mortality {}: {}'.format(var, mortality_total))
            logger.diagnostic(' total npp {}: {}'.format(var, npp_total))
            logger.diagnostic(' total grazed {}: {}'.format(var, grazing_total))

        for var in self.surface_out:
            logger.diagnostic(' mean {} surface concentration: {} mmol/m^3'.format(var, vs.npzd_tracers[var][vs.maskT[:, :, -1]].mean()))

        for var in self.bottom_out:
            logger.diagnostic(' mean {} bottom concentration: {} mmol/m^3'.format(var, vs.npzd_tracers[var][vs.bottom_mask].mean()))

    def read_restart(self, vs):
        pass

    def write_restart(self, vs, outfile):
        pass



try:
    import veros  # noqa: F401
except ImportError:
    raise RuntimeError(
        'veros-bgc needs Veros to be installed (try `pip install veros`)'
    )

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from veros_bgc.variables import MAIN_VARIABLES, CONDITIONAL_VARIABLES
from veros_bgc.settings import SETTINGS
from veros_bgc.core.npzd import npzd, setupNPZD
from veros_bgc.diagnostics.npzd_monitor import NPZDMonitor


__VEROS_INTERFACE__ = dict(
    name='biogeochemistry',
    setup_entrypoint=setupNPZD,
    run_entrypoint=npzd,
    settings=SETTINGS,
    variables=MAIN_VARIABLES,
    conditional_variables=CONDITIONAL_VARIABLES,
    diagnostics=[NPZDMonitor],
)



from collections import OrderedDict

from veros.settings import Setting


SETTINGS = OrderedDict([
    ('enable_npzd', Setting(False, bool, '')),

    # I don't like keeping these in settings, but I can't make them in variables
    # maybe keep selected rules
    ('recycled', Setting({}, dict, 'Amount of recycled material [mmol/m^3] for NPZD tracers')),
    ('mortality', Setting({}, dict, 'Amount of dead plankton [mmol/m^3] by species')),
    ('net_primary_production', Setting({}, dict, 'Primary production for each producing plankton species')),
    ('plankton_growth_functions', Setting({}, dict, 'Collection of functions calculating growth for plankton by species')),
    ('limiting_functions', Setting({}, dict, 'Collection of functions calculating limits to growth for plankton by species')),
    ('npzd_tracers', Setting({}, dict, 'Dictionary whose values point to veros variables for npzd tracers')),
    ('npzd_rules', Setting([], list, 'List of active rules in primary loop of BGC')),
    ('npzd_pre_rules', Setting([], list, 'List of rules to executed in the pre loop of BGC')),
    ('npzd_post_rules', Setting([], list, 'Rules to be executed after primary bio loop')),
    ('npzd_available_rules', Setting({}, dict, 'Every rule created is stored here, can be individual rules or collections of rules')),
    ('npzd_selected_rule_names', Setting([], list, 'name of selected rules')),
    ('npzd_export', Setting({}, dict, 'Exported material from npzd tracers by sinking')),
    ('npzd_import', Setting({}, dict, 'Imported material from npzd tracers from layer above. Takes same value as npzd_export scaled by level differences. Sea surface is 0')),
    ('zprefs', Setting({}, dict, 'Preference for zooplankton to graze on named tracers')),
    ('npzd_transported_tracers', Setting([], list, 'List of NPZD tracers which are transported')),
    ('npzd_advection_derivatives', Setting({}, dict, 'Stores derivates of advection term for tracers')),
    ('temporary_tracers', Setting({}, dict, 'Temporary copy of npzd_tracers for biogeochemistry loop')),

    ('light_attenuation_phytoplankton', Setting(0.047, float, 'Light attenuation of phytoplankton')),
    ('light_attenuation_water', Setting(0.04, float, 'Light attenuation of water [1/m]')),
    ('light_attenuation_ice', Setting(5.0, float, 'Light attenuation of ice [1/m]')),
    ('remineralization_rate_detritus', Setting(0, float, 'Remineralization rate of detritus [1/sec]')),
    ('bbio', Setting(0, float, 'the b in b ** (c*T)')),
    ('cbio', Setting(0, float, 'the c in b ** (c*T)')),
    ('maximum_growth_rate_phyto', Setting(0.0, float, 'Maximum growth rate parameter for phytoplankton in [1/sec]')),
    ('maximum_grazing_rate', Setting(0, float, 'Maximum grazing rate at 0 deg C [1/sec]')),
    ('fast_recycling_rate_phytoplankton', Setting(0, float, 'Fast-recycling mortality rate of phytoplankton [1/sec]')),
    ('saturation_constant_N', Setting(0.7, float, 'Half saturation constant for N uptake [mmol N / m^3]')),
    ('saturation_constant_Z_grazing', Setting(0.15, float, 'Half saturation constant for Z grazing [mmol/m^3]')),
    ('specific_mortality_phytoplankton', Setting(0, float, 'Specific mortality rate of phytoplankton')),
    ('quadric_mortality_zooplankton', Setting(0, float, 'Quadric mortality rate of zooplankton [1/ (mmol N ^2 s)]')),
    ('assimilation_efficiency', Setting(0, float, 'Effiency with which ingested prey is converted growth in zooplankton, range: [0,1]')),
    ('zooplankton_growth_efficiency', Setting(0, float, 'Zooplankton growth efficiency, range: [0,1]')),
    ('wd0', Setting(0 / 86400, float, 'Sinking speed of detritus at surface [m/s]')),
    ('mwz', Setting(1000, float, 'Depth below which sinking speed of detritus remains constant [m]')),
    ('mw', Setting(0.02 / 86400, float, 'Increase in sinking speed with depth [1/sec]')),
    ('zprefP', Setting(1, float, 'Zooplankton preference for grazing on Phytoplankton')),
    ('zprefZ', Setting(1, float, 'Zooplankton preference for grazing on other zooplankton')),
    ('zprefDet', Setting(1, float, 'Zooplankton preference for grazing on detritus')),
    ('redfield_ratio_PN', Setting(1. / 16, float, 'Refield ratio for P/N')),
    ('redfield_ratio_CP', Setting(7.1 * 16, float, 'Refield ratio for C/P')),
    ('redfield_ratio_ON', Setting(10.6, float, 'Redfield ratio for O/N')),
    ('redfield_ratio_CN', Setting(7.1, float, 'Redfield ratio for C/N')),
    ('trcmin', Setting(1e-13, float, 'Minimum npzd tracer value')),
    ('u1_min', Setting(1e-6, float, 'Minimum u1 value for calculating avg J')),
    ('zooplankton_max_growth_temp', Setting(20.0, float, 'Temperature (C) for which zooplankton growth rate no longer grows with temperature')),
    ('capr', Setting(0.022, float, 'Carbonate to carbon production ratio')),
])



"""
Classes for npzd tracers
"""
import numpy as np

from veros import veros_method

class NPZD_tracer(np.ndarray):
    """ Class for npzd tracers to store additional information about themselves.

    Note
    ----
    Inhenrits from numpy.ndarray to make it work seamless with array operations

    Parameters
    ----------
    input_array : :obj:`numpy.ndarray`
        Numpy array backing data

    name : :obj:`str`
        Identifier for the tracer, which must be unique within a given configuration

    sinking_speed : :obj:`numpy.ndarray`, optional
        Numpy array for how fast the tracer sinks in each cell

    transport : :obj:`bool` = True, optional
        Whether or not to include the tracer in physical transport

    light_attenuation : :obj:`numpy.ndarray`, optional
        Factor for how much light is blocked

    Attributes
    ----------
    name
        Identifier for the tracer, which must be unique within a given configuration

    description
        Description of the tracer represented by the class

    transport
        Whether or not to include the tracer in physical transport

    sinking_speed : :obj:`numpy.ndarray`, optional
        If set: how fast the tracer sinks in each cell

    light_attenuation : :obj:`numpy.ndarray`, optional
        If set: Factor for how much light is blocked
    """


    def __new__(cls, input_array, name, sinking_speed=None, light_attenuation=None, transport=True,
                description = None):
        obj = np.asarray(input_array).view(cls)
        if sinking_speed is not None:
            obj.sinking_speed = sinking_speed
        if light_attenuation is not None:
            obj.light_attenuation = light_attenuation

        obj.name = name
        obj.transport = transport

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

        # If we are slicing, obj will have __dir__ therefore we need to set attributes
        # on new sliced array
        if hasattr(obj, "__dir__"):
            for attribute in (set(dir(obj)) - set(dir(self))):
                setattr(self, attribute, getattr(obj, attribute))


class Recyclable_tracer(NPZD_tracer):
    """ A recyclable tracer

    This would be tracer, which may be a tracer like detritus, which can be recycled

    Parameters
    ----------
    input_array : :obj:`numpy.ndarray`
        Numpy array backing data

    name : :obj:`str`
        Identifier for the tracer, which must be unique within a given configuration

    recycling_rate
        A factor scaling the recycling by the population size

    **kwargs
        All named parameters accepted by super class

    Attributes
    ----------
    recycling_rate
        A factor scaling the recycling by the population size

    + All attributes held by super class
    """

    def __new__(cls, input_array, name, recycling_rate=0, **kwargs):
        obj = super().__new__(cls, input_array, name, **kwargs)
        obj.recycling_rate = recycling_rate

        return obj

    @veros_method(inline=True)
    def recycle(self, vs):
        """
        Recycling is temperature dependant by :obj:`vs.bct`
        """
        return vs.bct * self.recycling_rate * self


class Plankton(Recyclable_tracer):
    """ Class for plankton object, which is both recyclable and displays mortality

    This class is intended as a base for phytoplankton and zooplankton and not
    as a standalone class

    Note
    ----
    Typically, it would desirable to also set light attenuation


    Parameters
    ----------
    input_array : :obj:`numpy.ndarray`
        Numpy array backing data

    name : :obj:`str`
        Identifier for the tracer, which must be unique within a given configuration

    mortality_rate
        Rate at which the tracer is dying in mortality method

    **kwargs
        All named parameters accepted by super class

    Attributes
    ----------
    mortality_rate
        Rate at which the tracer is dying in mortality method

    + All attributes held by super class
    """

    def __new__(cls, input_array, name, mortality_rate=0, **kwargs):
        obj = super().__new__(cls, input_array, name, **kwargs)
        obj.mortality_rate = mortality_rate
        return obj

    @veros_method(inline=True)
    def mortality(self, vs):
        """
        The mortality rate scales linearly with population size
        """
        return self * self.mortality_rate


class Phytoplankton(Plankton):
    """ Phytoplankton also has primary production

    Parameters
    ----------
    input_array : :obj:`numpy.ndarray`
        Numpy array backing data

    name : :obj:`str`
        Identifier for the tracer, which must be unique within a given configuration

    growth_parameter
        Scaling factor for maximum potential growth

    **kwargs
        All named parameters accepted by super class

    Attributes
    ----------
    growth_parameter
        Scaling factor for maximum potential growth

    + All attributes held by super class
    """

    def __new__(cls, input_array, name, growth_parameter=0, **kwargs):
        obj = super().__new__(cls, input_array, name, **kwargs)
        obj.growth_parameter = growth_parameter

        return obj

    @veros_method(inline=True)
    def potential_growth(self, vs, grid_light, light_attenuation):
        """ Light limited growth, not limited growth """
        f1 = np.exp(-light_attenuation)  # available light
        jmax = self.growth_parameter * vs.bct  # maximum growth
        gd = jmax * vs.dayfrac[np.newaxis, :, np.newaxis]  # growth in fraction of day
        avej = self._avg_J(vs, f1, gd, grid_light, light_attenuation)  # light limited growth

        return jmax, avej

    @veros_method(inline=True)
    def _avg_J(self, vs, f1, gd, grid_light, light_attenuation):
        """ Average light over a triuneral cycle

        Note
        ----
        This calculation is only valid if grid_light / gd < 20
        """
        u1 = np.maximum(grid_light / gd, vs.u1_min)
        u2 = u1 * f1

        # NOTE: There is an approximation here: u1 < 20
        phi1 = np.log(u1 + np.sqrt(1 + u1**2)) - (np.sqrt(1 + u1**2) - 1) / u1
        phi2 = np.log(u2 + np.sqrt(1 + u2**2)) - (np.sqrt(1 + u2**2) - 1) / u2

        return gd * (phi1 - phi2) / light_attenuation


class Zooplankton(Plankton):
    """ Zooplankton displays quadratic mortality rate but otherwise is similar to ordinary phytoplankton

    Parameters
    ----------
    input_array : :obj:`numpy.ndarray`
        Numpy array backing data

    name : :obj:`str`
        Identifier for the tracer, which must be unique within a given configuration

    max_grazing
        Scaling factor for maximum grazing rate

    grazing_saturation_constant
        Saturation in Michaelis-Menten

    grazing_preferences
        Dictionary of preferences for grazing on other tracers

    assimilation_efficiency
        Fraction of grazed material ingested

    growth_efficiency
        Fraction of ingested material resulting in growth

    maximum_growth_temperature : = 20
        Temperature in Celsius where increasing temperature no longer increases grazing

    **kwargs
        All named parameters accepted by super class

    Attributes
    ----------
    max_grazing
        Scaling factor for maximum grazing rate

    grazing_saturation_constant
        Saturation in Michaelis-Menten

    grazing_preferences
        Dictionary of preferences for grazing on other tracers

    assimilation_efficiency
        Fraction of grazed material ingested

    growth_efficiency
        Fraction of ingested material resulting in growth

    maximum_growth_temperature
        Temperature in Celsius where increasing temperature no longer increases grazing

    + All attributes held by super class
    """

    def __new__(cls, input_array, name, max_grazing=0, grazing_saturation_constant=1,
                grazing_preferences={}, assimilation_efficiency=0,
                growth_efficiency=0,
                maximum_growth_temperature=20, **kwargs):
        obj = super().__new__(cls, input_array, name, **kwargs)

        obj.max_grazing = max_grazing
        obj.grazing_saturation_constant = grazing_saturation_constant
        obj.grazing_preferences = grazing_preferences
        obj.assimilation_efficiency = assimilation_efficiency
        obj.growth_efficiency = growth_efficiency
        obj.maximum_growth_temperature = maximum_growth_temperature
        obj._gmax = 0  # should be private

        return obj

    @veros_method(inline=True)
    def update_internal(self, vs):
        """
        Updates internal numbers, which are calculated only from Veros values
        """
        self._gmax = self.max_grazing * vs.bbio ** (vs.cbio *
                     np.minimum(self.maximum_growth_temperature, vs.temp[..., vs.tau]))

    @veros_method(inline=True)
    def mortality(self, vs):
        """
        Zooplankton is modelled with a quadratic mortality
        """
        return self.mortality_rate * self ** 2

    @veros_method(inline=True)
    def grazing(self, vs, tracers, flags):
        """
        Zooplankton grazing on set preys

        Parameters
        ----------
        tracers : dict
            Alle tracers, which can be grazed upon, not the same as grazing preferences

        flags : dict
            Flags for all tracers indicating whether they have been depleted within
            the current time step.

        Returns
        -------
        Dictionaries for grazing, digestion, excretion and sloppy feeding.
        Each grazed species is contained in the key corresponding to its name

        Note
        ----
        The result of this method is primarily useful in rules

        Note
        ----
        thetaZ is scaled by vs.redfield_ratio_PN. This may not be desirable in the general case
        """

        thetaZ = sum([pref_score * tracers[preference] for preference, pref_score
                      in self.grazing_preferences.items()])\
                 + vs.saturation_constant_Z_grazing * vs.redfield_ratio_PN

        ingestion = {preference: pref_score / thetaZ for preference, pref_score in self.grazing_preferences.items()}

        grazing = {preference: flags[preference] * flags[self.name] * self._gmax *
                   ingestion[preference] * tracers[preference] * self
                   for preference in ingestion}

        digestion = {preference: self.assimilation_efficiency * amount_grazed
                     for preference, amount_grazed in grazing.items()}

        excretion = {preference: (1 - self.growth_efficiency) * amount_digested
                     for preference, amount_digested in digestion.items()}

        sloppy_feeding = {preference: grazing[preference] - digestion[preference]
                          for preference in grazing}

        return grazing, digestion, excretion, sloppy_feeding



"""
Collection of rules to be used by the npzd module

Rules should always take 3 arguments:
    1. vs
    2. Name of source tracer
    3. Name of sink tracer
"""
from veros import veros_method

from . import atmospherefluxes


@veros_method(inline=True)
def empty_rule(*args):
    """An empty rule for providing structure"""
    return {}


@veros_method(inline=True)
def primary_production(vs, nutrient, plankton):
    """Primary production: Growth by consumption of light and nutrients"""
    return {nutrient: - vs.redfield_ratio_PN * vs.net_primary_production[plankton], plankton: vs.net_primary_production[plankton]}


@veros_method(inline=True)
def recycling(vs, plankton, nutrient, ratio):
    """Plankton or detritus is recycled into nutrients

    Parameters
    ----------
    ratio
        Factor for adjusting nutrient contribution. Typically a Redfield ratio.
    """
    return {plankton: - vs.recycled[plankton], nutrient: ratio * vs.recycled[plankton]}


@veros_method(inline=True)
def mortality(vs, plankton, detritus):
    """All dead matter from plankton is converted to detritus"""
    return {plankton: - vs.mortality[plankton], detritus: vs.mortality[plankton]}


@veros_method(inline=True)
def sloppy_feeding(vs, zooplankton, detritus):
    """When zooplankton graces, some is not eaten. This is converted to detritus."""
    sloppy_sum = sum(vs.sloppy_feeding.values())
    return {zooplankton: -sloppy_sum, detritus: sloppy_sum}


@veros_method(inline=True)
def grazing(vs, eaten, zooplankton):
    """Zooplankton grows by amount digested, eaten decreases by amount grazed"""
    return {eaten: - vs.grazing[eaten], zooplankton: vs.grazing[eaten]}


@veros_method(inline=True)
def excretion(vs, zooplankton, nutrient):
    """Zooplankton excretes nutrients after eating. Fecal matter, breathing..."""
    return {zooplankton: - vs.excretion_total, nutrient: vs.redfield_ratio_PN * vs.excretion_total}


@veros_method(inline=True)
def primary_production_from_DIC(vs, nutrient, plankton):
    """Stub rule for primary production consuming DIC. Should be replaced by common sink rule"""
    return {nutrient: - vs.redfield_ratio_CN * vs.net_primary_production[plankton]}


@veros_method(inline=True)
def recycling_to_po4(vs, plankton, phosphate):
    """Recycling to phosphate is scaled by redfield ratio P to N"""
    return recycling(vs, plankton, phosphate, vs.redfield_ratio_PN)


@veros_method(inline=True)
def recycling_to_no3(vs, plankton, no3):
    """Recycling to nitrate needs no scaling"""
    return {no3: recycling(vs, plankton, no3, 1)[no3]}


@veros_method(inline=True)
def recycling_to_dic(vs, plankton, dic):
    """Recycling to carbon

    Is scaled by redfield ratio C to N removing plankton has already been done by po4
    """
    return {dic: recycling(vs, plankton, dic, vs.redfield_ratio_CN)[dic]}


@veros_method(inline=True)
def recycling_phyto_to_dic(vs, plankton, dic):
    """Fast recycling of phytoplankton to DIC"""
    return {dic: recycling(vs, plankton, dic, vs.redfield_ratio_CN)[dic]}


@veros_method(inline=True)
def excretion_dic(vs, zooplankton, nutrient):
    """Zooplankton excretes nutrients after eating. Poop, breathing..."""
    return {nutrient: vs.redfield_ratio_CN * vs.excretion_total}


@veros_method(inline=True)
def calcite_production(vs, plankton, DIC, calcite):
    """Calcite is produced at a rate similar to detritus

    Intended for use with a smoothing rule
    If explicit tracking of calcite is desired use
    rules for the explicit relationship
    """

    # changes to production of calcite
    dprca = (vs.mortality[plankton] + vs.grazing[plankton] *
             (1 - vs.assimilation_efficiency)) * vs.capr * vs.redfield_ratio_CN

    return {DIC: -dprca, calcite: dprca}


@veros_method(inline=True)
def calcite_production_alk(vs, plankton, alkalinity, calcite):
    """alkalinity is scaled by DIC consumption"""
    return {alkalinity: 2 * calcite_production(vs, plankton, 'DIC', calcite)['DIC']}


@veros_method(inline=True)
def calcite_production_phyto(vs, DIC, calcite):
    """DIC is consumed to produce calcite.

    How much depends on the mortality and sloppy feeding of and on phytoplankton and zooplankton
    """
    return calcite_production(vs, 'phytoplankton', DIC, calcite)


@veros_method(inline=True)
def calcite_production_phyto_alk(vs, alkalinity, calcite):
    """Scaled version off calcite_production_phyto for alkalinity"""
    return calcite_production_alk(vs, 'phytoplankton', alkalinity, calcite)


@veros_method(inline=True)
def post_redistribute_calcite(vs, calcite, tracer):
    """Post rule to redistribute produced calcite"""
    total_production = (vs.temporary_tracers[calcite] * vs.dzt).sum(axis=2)
    redistributed_production = total_production[:, :, np.newaxis] * vs.rcak
    return {tracer: redistributed_production}


@veros_method(inline=True)
def post_redistribute_calcite_alk(vs, calcite, alkalinity):
    """Post rule reusing post_redistribute_calcite to redistribute alkalinity"""
    return {alkalinity: 2 * post_redistribute_calcite(vs, calcite, 'DIC')['DIC']}


@veros_method(inline=True)
def pre_reset_calcite(vs, tracer, calcite):
    """Pre rule to reset calcite production"""
    return {calcite: - vs.temporary_tracers[calcite]}


@veros_method
def co2_surface_flux(vs, co2, dic):
    """Pre rule to add or remove DIC from surface layer"""
    atmospherefluxes.carbon_flux(vs)
    flux = vs.cflux * vs.dt_tracer / vs.dzt[-1]
    return {dic: flux}  # NOTE we don't have an atmosphere, so this rules is a stub


@veros_method(inline=True)
def dic_alk_scale(vs, DIC, alkalinity):
    """Redistribute change in DIC as change in alkalinity"""
    return {alkalinity: (vs.temporary_tracers[DIC] - vs.npzd_tracers[DIC][:, :, :, vs.tau]) / vs.redfield_ratio_CN}


@veros_method(inline=True)
def bottom_remineralization(vs, source, sink, scale):
    """Exported material falling through the ocean floor is converted to nutrients

    Note
    ----
    There can be no source, because that is handled by the sinking code

    Parameters
    ----------
    scale
        Factor to convert remineralized material to nutrient. Typically Redfield ratio.
    """
    return {sink: vs.npzd_export[source][vs.bottom_mask] * scale}


@veros_method(inline=True)
def bottom_remineralization_detritus_po4(vs, detritus, po4):
    """Remineralize detritus at the bottom to PO4"""
    return bottom_remineralization(vs, detritus, po4, vs.redfield_ratio_PN)


@veros_method(inline=True)
def bottom_remineralization_detritus_DIC(vs, detritus, DIC):
    """Remineralize detritus at the bottom to DIC"""
    return bottom_remineralization(vs, detritus, DIC, vs.redfield_ratio_CN)



#!/usr/bin/env python

import os

import h5netcdf
import ruamel.yaml as yaml

from veros import VerosSetup, veros_method, distributed
from veros.variables import Variable
import veros.tools

import veros_bgc

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
DATA_FILES = veros.tools.get_assets(
    'global_4deg',
    os.path.join(BASE_PATH, 'assets.yml')
)


class GlobalFourDegreeBGC(VerosSetup):
    """Global 4 degree model with 15 vertical levels and biogeochemistry.
    """
    __veros_plugins__ = (veros_bgc,)

    @veros_method
    def set_parameter(self, vs):
        vs.identifier = '4deg'

        vs.nx, vs.ny, vs.nz = 90, 40, 15
        vs.dt_mom = 1800.0
        vs.dt_tracer = 86400.0
        vs.dt_bio = vs.dt_tracer // 4
        vs.runlen = 0.

        vs.trcmin = 0
        vs.enable_npzd = True
        vs.enable_carbon = True

        with open(os.path.join(BASE_PATH, 'npzd.yml')) as yaml_file:
            cfg = yaml.safe_load(yaml_file)['npzd']
            vs.npzd_selected_rules = cfg['selected_rules']

        vs.remineralization_rate_detritus = 0.09 / 86400
        vs.bbio = 1.038
        vs.cbio = 1.0
        vs.maximum_growth_rate_phyto = 0.23 / 86400
        vs.maximum_grazing_rate = 0.13 / 86400
        vs.fast_recycling_rate_phytoplankton = 0.025 / 86400
        vs.specific_mortality_phytoplankton = 0.035 / 86400
        vs.quadric_mortality_zooplankton = 0.06 / 86400
        vs.zooplankton_growth_efficiency = 0.60
        vs.assimilation_efficiency = 0.5
        vs.wd0 = 2 / 86400
        vs.mwz = 1000
        vs.mw = 0.02 / 86400
        vs.dcaco3 = 2500

        vs.coord_degree = True
        vs.enable_cyclic_x = True

        vs.congr_epsilon = 1e-8
        vs.congr_max_iterations = 20000

        vs.enable_neutral_diffusion = True
        vs.K_iso_0 = 1000.0
        vs.K_iso_steep = 500.0
        vs.iso_dslope = 0.001
        vs.iso_slopec = 0.001
        vs.enable_skew_diffusion = True

        vs.enable_hor_friction = True
        vs.A_h = (4 * vs.degtom)**3 * 2e-11
        vs.enable_hor_friction_cos_scaling = True
        vs.hor_friction_cosPower = 1

        vs.enable_implicit_vert_friction = True
        vs.enable_tke = True
        vs.c_k = 0.1
        vs.c_eps = 0.7
        vs.alpha_tke = 30.0
        vs.mxl_min = 1e-8
        vs.tke_mxl_choice = 2
        vs.kappaM_min = 2e-4
        vs.kappaH_min = 7e-5  # usually 2e-5
        vs.enable_Prandtl_tke = False
        vs.enable_kappaH_profile = True

        # eke
        vs.K_gm_0 = 1000.0
        vs.enable_eke = False
        vs.eke_k_max = 1e4
        vs.eke_c_k = 0.4
        vs.eke_c_eps = 0.5
        vs.eke_cross = 2.
        vs.eke_crhin = 1.0
        vs.eke_lmin = 100.0
        vs.enable_eke_superbee_advection = False
        vs.enable_eke_isopycnal_diffusion = False

        # idemix
        vs.enable_idemix = False
        vs.enable_idemix_hor_diffusion = False
        vs.enable_eke_diss_surfbot = False
        vs.eke_diss_surfbot_frac = 0.2  #  fraction which goes into bottom
        vs.enable_idemix_superbee_advection = False

        vs.eq_of_state_type = 5

        # custom variables
        vs.nmonths = 12
        vs.variables.update(
            sss_clim=Variable('sss_clim', ('xt', 'yt', 'nmonths'), '', '', time_dependent=False),
            sst_clim=Variable('sst_clim', ('xt', 'yt', 'nmonths'), '', '', time_dependent=False),
            qnec=Variable('qnec', ('xt', 'yt', 'nmonths'), '', '', time_dependent=False),
            qnet=Variable('qnet', ('xt', 'yt', 'nmonths'), '', '', time_dependent=False),
            taux=Variable('taux', ('xt', 'yt', 'nmonths'), '', '', time_dependent=False),
            tauy=Variable('tauy', ('xt', 'yt', 'nmonths'), '', '', time_dependent=False),
        )

    @veros_method
    def _read_forcing(self, vs, var):
        with h5netcdf.File(DATA_FILES['forcing'], 'r') as infile:
            var_obj = infile.variables[var]
            return np.array(var_obj, dtype=str(var_obj.dtype)).T

    @veros_method
    def set_grid(self, vs):
        ddz = np.array([50., 70., 100., 140., 190., 240., 290., 340.,
                        390., 440., 490., 540., 590., 640., 690.])
        vs.dzt[:] = ddz[::-1]
        vs.dxt[:] = 4.0
        vs.dyt[:] = 4.0
        vs.y_origin = -78.0
        vs.x_origin = 4.0

        idx_global, _ = distributed.get_chunk_slices(vs, ('xt', 'yt', 'zt'))

        with h5netcdf.File(DATA_FILES['corev2'], 'r') as infile:
            vs.swr_initial = infile.variables['SWDN_MOD'][idx_global[::-1]].T

    @veros_method
    def set_coriolis(self, vs):
        vs.coriolis_t[...] = 2 * vs.omega * np.sin(vs.yt[np.newaxis, :] / 180. * vs.pi)

    @veros_method(dist_safe=False, local_variables=[
        'kbot'
    ])
    def set_topography(self, vs):
        bathymetry_data = self._read_forcing(vs, 'bathymetry')
        salt_data = self._read_forcing(vs, 'salinity')[:, :, ::-1]
        mask_salt = salt_data == 0.
        vs.kbot[2:-2, 2:-2] = 1 + np.sum(mask_salt.astype(np.int), axis=2)
        mask_bathy = bathymetry_data == 0
        vs.kbot[2:-2, 2:-2][mask_bathy] = 0
        vs.kbot[vs.kbot == vs.nz] = 0

    @veros_method(dist_safe=False, local_variables=[
        'taux', 'tauy', 'qnec', 'qnet', 'sss_clim', 'sst_clim',
        'temp', 'salt', 'taux', 'tauy', 'area_t', 'maskT',
        'forc_iw_bottom', 'forc_iw_surface', 'zw', 'maskT',
        'phytoplankton', 'zooplankton', 'detritus', 'po4',
        'dic', 'atmospheric_co2', 'alkalinity', 'hSWS'
    ])
    def set_initial_conditions(self, vs):
        # initial conditions for T and S
        temp_data = self._read_forcing(vs, 'temperature')[:, :, ::-1]
        vs.temp[2:-2, 2:-2, :, :2] = temp_data[:, :, :, np.newaxis] * \
            vs.maskT[2:-2, 2:-2, :, np.newaxis]

        salt_data = self._read_forcing(vs, 'salinity')[:, :, ::-1]
        vs.salt[2:-2, 2:-2, :, :2] = salt_data[..., np.newaxis] * vs.maskT[2:-2, 2:-2, :, np.newaxis]

        # use Trenberth wind stress from MITgcm instead of ECMWF (also contained in ecmwf_4deg.cdf)
        vs.taux[2:-2, 2:-2, :] = self._read_forcing(vs, 'tau_x')
        vs.tauy[2:-2, 2:-2, :] = self._read_forcing(vs, 'tau_y')

        # heat flux
        with h5netcdf.File(DATA_FILES['ecmwf'], 'r') as ecmwf_data:
            qnec_var = ecmwf_data.variables['Q3']
            vs.qnec[2:-2, 2:-2, :] = np.array(qnec_var, dtype=str(qnec_var.dtype)).transpose()
            vs.qnec[vs.qnec <= -1e10] = 0.0

        q = self._read_forcing(vs, 'q_net')
        vs.qnet[2:-2, 2:-2, :] = -q
        vs.qnet[vs.qnet <= -1e10] = 0.0

        fxa = np.sum(vs.qnet[2:-2, 2:-2, :] * vs.area_t[2:-2, 2:-2, np.newaxis]) \
              / 12 / np.sum(vs.area_t[2:-2, 2:-2])
        print(' removing an annual mean heat flux imbalance of %e W/m^2' % fxa)
        vs.qnet[...] = (vs.qnet - fxa) * vs.maskT[:, :, -1, np.newaxis]

        # SST and SSS
        vs.sst_clim[2:-2, 2:-2, :] = self._read_forcing(vs, 'sst')
        vs.sss_clim[2:-2, 2:-2, :] = self._read_forcing(vs, 'sss')

        if vs.enable_idemix:
            vs.forc_iw_bottom[2:-2, 2:-2] = self._read_forcing(vs, 'tidal_energy') / vs.rho_0
            vs.forc_iw_surface[2:-2, 2:-2] = self._read_forcing(vs, 'wind_energy') / vs.rho_0 * 0.2

        if vs.enable_npzd:
            phytoplankton = 0.14 * np.exp(vs.zw / 100) * vs.maskT
            zooplankton = 0.014 * np.exp(vs.zw / 100) * vs.maskT

            vs.phytoplankton[:, :, :, :] = phytoplankton[..., np.newaxis]
            vs.zooplankton[:, :, :, :] = zooplankton[..., np.newaxis]
            vs.detritus[:, :, :, :] = 1e-4 * vs.maskT[..., np.newaxis]

            vs.po4[:, :, :, :] = 2.2
            vs.po4[...] *= vs.maskT[..., np.newaxis]

        if vs.enable_carbon:
            vs.dic[...] = 2300 * vs.maskT[..., np.newaxis]
            vs.atmospheric_co2[...] = 280
            vs.alkalinity[...] = 2400 * vs.maskT[..., np.newaxis]

            vs.hSWS[...] = 5e-7

    @veros_method
    def set_forcing(self, vs):
        year_in_seconds = 360 * 86400.
        (n1, f1), (n2, f2) = veros.tools.get_periodic_interval(
            vs.time, year_in_seconds, year_in_seconds / 12., 12
        )

        # wind stress
        vs.surface_taux[...] = (f1 * vs.taux[:, :, n1] + f2 * vs.taux[:, :, n2])
        vs.surface_tauy[...] = (f1 * vs.tauy[:, :, n1] + f2 * vs.tauy[:, :, n2])

        # tke flux
        if vs.enable_tke:
            vs.forc_tke_surface[1:-1, 1:-1] = np.sqrt((0.5 * (vs.surface_taux[1:-1, 1:-1] \
                                                                + vs.surface_taux[:-2, 1:-1]) / vs.rho_0)**2
                                                      + (0.5 * (vs.surface_tauy[1:-1, 1:-1] \
                                                                + vs.surface_tauy[1:-1, :-2]) / vs.rho_0)**2)**(3. / 2.)
        # heat flux : W/m^2 K kg/J m^3/kg = K m/s
        cp_0 = 3991.86795711963
        sst = f1 * vs.sst_clim[:, :, n1] + f2 * vs.sst_clim[:, :, n2]
        qnec = f1 * vs.qnec[:, :, n1] + f2 * vs.qnec[:, :, n2]
        qnet = f1 * vs.qnet[:, :, n1] + f2 * vs.qnet[:, :, n2]
        vs.forc_temp_surface[...] = (qnet + qnec * (sst - vs.temp[:, :, -1, vs.tau])) \
                                       * vs.maskT[:, :, -1] / cp_0 / vs.rho_0

        # salinity restoring
        t_rest = 30 * 86400.0
        sss = f1 * vs.sss_clim[:, :, n1] + f2 * vs.sss_clim[:, :, n2]

        vs.forc_salt_surface[:] = 1. / t_rest * \
            (sss - vs.salt[:, :, -1, vs.tau]) * vs.maskT[:, :, -1] * vs.dzt[-1]

        # apply simple ice mask
        mask = np.logical_and(vs.temp[:, :, -1, vs.tau] * vs.maskT[:, :, -1] < -1.8,
                              vs.forc_temp_surface < 0.)
        vs.forc_temp_surface[mask] = 0.0
        vs.forc_salt_surface[mask] = 0.0

        if vs.enable_npzd:
            # incoming shortwave radiation for plankton production
            vs.swr[2:-2, 2:-2] = (f1 * vs.swr_initial[:, :, n1] + f2 * vs.swr_initial[:, :, n2])

    @veros_method
    def set_diagnostics(self, vs):
        vs.diagnostics['snapshot'].output_frequency = 360 * 86400.

        vs.diagnostics['overturning'].output_frequency = 360 * 86400.
        vs.diagnostics['overturning'].sampling_frequency = vs.dt_tracer

        vs.diagnostics['energy'].output_frequency = 360 * 86400.
        vs.diagnostics['energy'].sampling_frequency = 86400

        average_vars = [
            'phytoplankton', 'po4', 'zooplankton', 'detritus', 'wind_speed', 'dic', 'alkalinity',
            'temp', 'salt', 'u', 'v', 'surface_tauy', 'surface_taux', 'kappaH', 'psi', 'rho',
            'pCO2', 'dpCO2', 'dco2star', 'cflux', 'atmospheric_co2'
        ]
        vs.diagnostics['averages'].output_variables = average_vars
        vs.diagnostics['averages'].output_frequency = 360 * 86400.
        vs.diagnostics['averages'].sampling_frequency = 86400

        vs.diagnostics['npzd'].output_frequency = 10 * 86400
        vs.diagnostics['npzd'].save_graph = False

    @veros_method
    def after_timestep(self, vs):
        pass


@veros.tools.cli
def run(*args, **kwargs):
    simulation = GlobalFourDegreeBGC(*args, **kwargs)
    simulation.setup()
    simulation.run()


if __name__ == '__main__':
    run()



"""
Functions calculating atmosphere-ocean fluxes

"""
from veros import veros_method, runtime_settings as rs
from veros.core import utilities


@veros_method
def carbon_flux(vs):
    """Calculates flux of CO2 over the ocean-atmosphere boundary

    This is an adaptation of co2_calc_SWS from UVic ESCM

    Note
    ----
    This was written without an atmosphere component in veros.
    Therefore an atmospheric pressure of 1 atm is assumed.
    The concentration of CO2 (in units of ppmv) may be set in vs.atmospheric_co2

    Note
    ----
    This was written without an explicit sea ice component. Therefore a full
    ice cover is assumed when temperature is below -1.8C and temperature forcing is negative

    Returns
    -------
    numpy.ndarray(vs.nx, vs.ny) with flux in units of :math:`mmol / m^2 / s`
    Positive indicates a flux into the ocean
    """

    icemask = np.logical_and(vs.temp[:, :, -1, vs.tau] * vs.maskT[:, :, -1] < -1.8,
            vs.forc_temp_surface < 0.0)
    ao = np.logical_not(icemask)

    atmospheric_pressure = 1  # atm  NOTE: We don't have an atmosphere yet, hence constant pressure

    # TODO get actual wind speed rather than deriving from wind stress
    wind_speed = np.sqrt(np.abs(vs.surface_taux / vs.rho_0) + np.abs(vs.surface_tauy / vs.rho_0)) * 500
    vs.wind_speed = wind_speed


    # xconv is constant to convert piston_vel from cm/hr -> cm/s
    # here it is 100.*a*xconv (100 => m to cm, a=0.337, xconv=1/3.6e+05)
    xconv = 0.337 / 3.6e5
    xconv *= 0.75  # NOTE: This seems like an approximation I don't know where they got it

    vs.dco2star = co2calc_SWS(vs, vs.temp[:, :, -1, vs.tau],  # [degree C]
                              vs.salt[:, :, -1, vs.tau],  # [g/kg]
                              vs.dic[:, :, -1, vs.tau] * 1e-3,  # [mmol -> mol]
                              vs.alkalinity[:, :, -1, vs.tau] * 1e-3,  # [mmol -> mol]
                              vs.atmospheric_co2,  # [ppmv]
                              atmospheric_pressure)  # atm

    # Schmidt number for CO2
    # Wanninkhof, 1992, table A
    scco2 = 2073.1 - 125.62 * vs.temp[:, :, -1, vs.tau] + 3.6276 * vs.temp[:, :, -1, vs.tau] ** 2 - 0.043219 * vs.temp[:, :, -1, vs.tau] ** 3

    piston_vel = ao * xconv * (wind_speed) ** 2 * ((scco2/660.0)**(-0.5))
    # NOTE: According to https://www.soest.hawaii.edu/oceanography/courses/OCN623/Spring%202015/Gas_Exchange_2015_one-lecture1.pdf there are 3 regimes we are looking at the wavy surface
    # NOTE: https://aslopubs.onlinelibrary.wiley.com/doi/pdf/10.4319/lom.2014.12.351 uses the correct form for scco2
    # Sweeney et al. 2007

    # 1e3 added to convert to mmol / m^2 / s
    co2_flux = piston_vel * vs.dco2star * vs.maskT[:, :, -1] * 1e3

    utilities.enforce_boundaries(vs, co2_flux)

    vs.cflux[...] = co2_flux
    return vs.cflux


@veros_method
def co2calc_SWS(vs, temperature, salinity, dic_in, ta_in, co2_in, atmospheric_pressure):
    """ Calculate delta co2*

        Calculate delta co2* from total alkalinty and total CO2 at temperature, salinity and atmosphere total pressure
        This function needs a better name

        Parameters
        ----------

        temperature
            sea surface temperature

        salinity
            sea surface salinity

        dic_in
            DIC in surface layer in units [mol/m^3]

        ta_in
            Total alkalinity in surface layer in units [mol/m^3]

        co2_in
            atmospheric co2

        atmospheric_pressure
            atmospheric pressure in [atm]

        Note
        ----
        It is assumed, that total Si and P are constant.
        Several other element concentration are approximated from temperature and salinity

        Returns
        -------

        $\delta CO_2\star$
    """

    sit_in = np.ones_like(temperature) * 7.6875e-3  # [mol/m^3] estimated total Si
    pt_in = np.ones_like(temperature) * 0.5125e-3  # [mol/m^3 ] estimated total P

    temperature_in_kelvin = temperature + 273.15

    # Change units from the input of mol/m^3 -> mol/kg:
    # where the ocean's mean surface density is 1024.5 kg/m^3
    # Note: mol/kg are actually what the body of this routine uses for calculations

    permil = 1.0 / 1024.5
    pt = pt_in * permil  # total P
    sit = sit_in * permil  # total Si
    ta = ta_in * permil  # total alkalinity
    dic = dic_in * permil

    # convert from uatm to atm, but it's in ppmv
    permeg = 1e-6
    co2 = co2_in * permeg

    scl = salinity / 1.80655  # convenience parameter
    _is = 19.924 * salinity / (1000.0 - 1.005 * salinity)  # ionic _strength

    # Concentrations for borate, sulfate, and flouride
    bt = 0.000232 * scl / 10.811  # Uppstrom (1974)
    st = 0.14 * scl / 96.062  # Morris & Riley (1966)
    ft = 0.000067 * scl / 18.9984  # Riley (1965)

    # Weiss & Price (1980, Mar. Chem., 8, 347-359; Eq 13 with table 6 values)
    ff = np.exp(-162.8301 + 218.2968 / (temperature_in_kelvin / 100)
                + 90.9241 * np.log(temperature_in_kelvin / 100)
                - 1.47696 * (temperature_in_kelvin / 100)**2
                + salinity * (0.025695 - 0.025225 * temperature_in_kelvin / 100
                + 0.0049867 * (temperature_in_kelvin / 100)**2))

    # K0 from Weiss 1974
    k0 = np.exp(93.4517 / (temperature_in_kelvin / 100) - 60.2409 + 23.3585
            * np.log(temperature_in_kelvin / 100) + salinity * (
                0.023517 - 0.023656 * temperature_in_kelvin / 100 + 0.0047036 * (temperature_in_kelvin / 100) ** 2))

    # Now calculate FugFac according to Weiss (1974) Marine Chemestry
    rt_x = 83.1451 * temperature_in_kelvin  # Gas constant times temperature
    delta_x = (57.7 - 0.118 * temperature_in_kelvin)  # In the text '\delta CO_2-air
    b_x = -1636.75 + 12.0408 * temperature_in_kelvin - 0.0327957 * temperature_in_kelvin ** 2 \
            + 3.16528 * 1e-5 * temperature_in_kelvin ** 3  # equation 6: Second viral coefficient B(T) (cm^3/mol)
    FugFac = np.exp((b_x + 2 * delta_x) / rt_x)  # equation 9 without front factor and ignoring pressure factors


    # k1 = [H][HCO3]/[H2CO3]
    # k2 = [H][CO3]/[HCO3]
    # Millero p.664 (1995) using Mehrbach et al. data on SEAWATER scale
    # (Original reference: Dickson and Millero, DSR, 1987)
    k1 = 10 ** (-1.0 * (3670.7 / temperature_in_kelvin - 62.088 + 9.7944 * np.log(temperature_in_kelvin)
                - 0.0118 * salinity + 0.000116 * salinity**2))

    k2 = 10 ** (-1 * (1394.7 / temperature_in_kelvin + 4.777 - 0.0184 * salinity + 0.000118 * salinity**2))

    # k1p = [H][H2PO4]/[H3PO4]
    # Millero p.670 (1995)
    k1p = np.exp(-4576.752 / temperature_in_kelvin + 115.540 - 18.453 * np.log(temperature_in_kelvin)
                 + (-106.736 / temperature_in_kelvin + 0.69171) * np.sqrt(salinity)
                 + (-0.65643 / temperature_in_kelvin - 0.01844) * salinity)

    # k2p = [H][HPO3]/[H2PO4]
    k2p = np.exp(-8814.715 / temperature_in_kelvin + 172.1033 - 27.927 * np.log(temperature_in_kelvin)
                 + (-160.340 / temperature_in_kelvin + 1.3566) * np.sqrt(salinity)
                 + (0.37335 / temperature_in_kelvin - 0.05778) * salinity)

    # k3p = [H][PO4]/[HPO4]
    k3p = np.exp(-3070.75 / temperature_in_kelvin - 18.126
                 + (17.27039 / temperature_in_kelvin + 2.81197) * np.sqrt(salinity)
                 + (-44.99486 / temperature_in_kelvin - 0.09984) * salinity)

    # ksi = [H][SiO(OH)3]/[Si(OH)4]
    # Millero p.671 (1995) using data from Yao and Millero (1995)
    # change to (mol/ kg soln)
    # depth dependancy assumed to be the same as boric acid
    # typo in Millero 1994 corrected in sign of 0.1622
    ksi = np.exp(-8904.2 / temperature_in_kelvin + 117.400 - 19.334*np.log(temperature_in_kelvin)
                 + (-458.79 / temperature_in_kelvin + 3.5913) * np.sqrt(_is)
                 + (188.74 / temperature_in_kelvin - 1.5998) * _is
                 + (-12.1652 / temperature_in_kelvin + 0.07871) * _is**2
                 + np.log(1.0 - 0.001005 * salinity))

    # kw = [H][OH]
    # Millero p.670 (1995) using composite data
    # pressure dependancy in Millero 1994 corrected for sea water from
    # Millero 1983

    kw = np.exp(-13847.26 / temperature_in_kelvin + 148.9802
                - 23.6521 * np.log(temperature_in_kelvin)
                + (118.67 / temperature_in_kelvin - 5.977
                   + 1.0495 * np.log(temperature_in_kelvin)) * np.sqrt(salinity) - 0.01615 * salinity)

    # ks = [H][SO4]/[HSO4] on free H scale
    # Dickson (1990, J. chem. Thermodynamics 22, 113)
    # change to (mol/ kg soln)
    ks = np.exp(-4276.1 / temperature_in_kelvin + 141.328 - 23.093 * np.log(temperature_in_kelvin)
                + (-13856 / temperature_in_kelvin + 324.57 - 47.986 * np.log(temperature_in_kelvin)) * np.sqrt(_is)
                + (35474 / temperature_in_kelvin - 771.54 + 114.723 * np.log(temperature_in_kelvin)) * _is
                - 2698 / temperature_in_kelvin * _is**1.5 + 1776 / temperature_in_kelvin * _is**2
                + np.log(1.0 - 0.001005 * salinity))

    # kf = [H][F]/[HF] on free H scale
    # Dickson and Riley (1979)
    # change to (mol/ kg soln)
    kf = np.exp(1590.2 / temperature_in_kelvin - 12.641 + 1.525 * np.sqrt(_is) + np.log(1.0 - 0.001005 * salinity))

    # kb = [H][BO2]/[HBO2]
    # Dickson p.673 (1990)
    # change from htotal to hSWS
    # typo in Millero 1994 corrected in sign of 0.1622
    kb = np.exp((-8966.90 - 2890.53 * np.sqrt(salinity) - 77.942 * salinity
                 + 1.728 * salinity**1.5 - 0.0996 * salinity**2) / temperature_in_kelvin
                 + (148.0248 + 137.1942 * np.sqrt(salinity) + 1.62142 * salinity)
                 + (-24.4344 - 25.085 * np.sqrt(salinity) - 0.2474 * salinity) * np.log(temperature_in_kelvin)
                 + 0.053105 * np.sqrt(salinity) * temperature_in_kelvin
                 + np.log((1 + (st / ks) + (ft / kf)) / (1 + (st / ks))))

    # From UVic ESCM comments
    # Calculate [H+] SWS when DIC and TA are known at T, S and 1 atm.
    # The solution converges to err of xacc. The solution must be within
    # the range x1 to x2.
    #
    # If DIC and TA are known then either a root finding or iterative method
    # must be used to calculate hSWS. In this case we use the Newton-Raphson
    # 'safe' method taken from 'Numerical Recipes' (function 'rtsafe.f' with
    # error trapping removed).

    xacc = 1e-10

    # NOTE hSWS should never exceed the safe values, but we can check it
    limit_min = 1e-8
    boundary = np.maximum(limit_min,
                          np.minimum(1e-6 - vs.hSWS, vs.hSWS - 1e-10))

    # boundary guesses are placed equidistant from the guessed best solution
    in1 = vs.hSWS + boundary  # input guess 1
    in2 = vs.hSWS - boundary  # input guess 2

    # where to run the optimization - skip land
    iter_mask = np.empty_like(in1, dtype=np.bool)
    iter_mask[...] = vs.maskT[:, :, -1]

    # Find hSWS by root finding algorithm
    vs.hSWS = drtsafe_masked(vs, iter_mask, ta_iter_SWS, in1, in2,
                             args=(k1, k2, k1p, k2p, k3p, st, ks, kf, ft,
                                   dic, ta, sit, ksi, pt, bt, kw, kb), accuracy=xacc)

    # Calculate [CO2*] as defined in DOE Methods Handbook 1993 Ver. 2,
    # ORNL/CDIC-74, Dickson and Goyet, eds. (Ch 2 p 1+, Eq A.49)
    co2star = dic * vs.hSWS**2 / (vs.hSWS**2 + k1 * vs.hSWS + k1 * k2)
    co2starair = co2 * ff * atmospheric_pressure
    dco2star = co2starair - co2star  # effective CO2 pressure difference

    # pCO2 and dpCO2 are for diagnostic purposes only
    pCO2 = co2star / (k0 * FugFac)  # partial CO2 pressure
    dpCO2 = pCO2 - co2 * atmospheric_pressure  # difference in CO2 pressure between ocean and atmosphere

    # Convert units back
    vs.co2star[:, :] = co2star / permil
    vs.dco2star[:, :] = dco2star / permil

    vs.pCO2[:, :] = pCO2 / permeg
    vs.dpCO2[:, :] = dpCO2 / permeg

    return vs.dco2star


@veros_method
def ta_iter_SWS_numpy(vs, x, k1, k2, k1p, k2p, k3p, st, ks, kf, ft, dic, ta, sit, ksi, pt, bt, kw, kb):
    """
    Sum of free protons calculated from saturation constants

    Parameters are same as ta_iter_SWS

    Note
    ----
    This function is internal and is intended to be called by ta_iter_SWS
    """
    x2 = x * x
    x3 = x2 * x
    k12 = k1 * k2
    k12p = k1p * k2p
    k123p = k12p * k3p
    c = 1.0 + st / ks + ft / kf
    a = x3 + k1p * x2 + k12p * x + k123p
    a2 = a * a
    da = 3.0 * x2 + 2.0 * k1p * x + k12p
    b = x2 + k1 * x + k12
    b2 = b * b
    db = 2.0 * x + k1

    # fn = hco3+co3+borate+oh+hpo4+2*po4+silicate-hfree-hso4-hf-h3po4-ta
    fn = k1 * x * dic / b + 2.0 * dic * k12 / b + \
         bt / (1.0 + x / kb) + kw / x + pt * k12p * x / a + \
         2.0 * pt * k123p / a + sit / (1.0 + x / ksi) - x / c - \
         st / (1.0 + ks / (x / c)) - ft / (1.0 + kf / (x / c)) -\
         pt * x3/a - ta

    # df = dfn/dx
    df = ((k1 * dic * b) - k1 * x * dic * db) / b2 - 2.0 * dic * k12 * db / b2 - \
         bt / kb / (1.0 + x / kb)**2.0 - kw / x2 + (pt * k12p * (a - x * da)) / a2 - \
         2.0 * pt * k123p * da / a2 - sit / ksi / (1.0 + x / ksi)**2 - 1.0 / c - \
         st * (1.0 + ks / (x / c))**(-2.0) * (ks * c / x2) - ft * (1.0 + kf / (x / c))**(-2.0) * \
         (kf * c / x2) - pt * x2 * (3.0 * a - x * da) / a2

    return fn, df


@veros_method
def ta_iter_SWS_bohrium(*args):
    """
    Sum of free proton calculated from saturation constants

    Parameters are same as ta_iter_SWS

    Note
    ----
    This function is internal. It was made explicitly for bohrium, making use of a mask to limit the number of calculations
    """

    arg_behaving = [np.user_kernel.make_behaving(args[1], dtype=np.int32)]

    for i, arg in enumerate(args[2:]):
        arg_behaving.append(np.user_kernel.make_behaving(arg, dtype=np.double))

    mask, x, k1, k2, k1p, k2p, k3p, st, ks, kf, ft, dic, ta, sit, ksi, pt, bt, kw, kb = arg_behaving

    fn = np.empty_like(x)
    df = np.empty_like(x)

    kernel = """
    #include <stdint.h>
    #include <stdlib.h>

    void execute(int *mask, double *x, double *k1, double *k2, double *k1p, double *k2p, double *k3p, double *st, double *ks, double *kf, double *ft, double *dic, double *ta, double *sit, double *ksi, double *pt, double *bt, double *kw, double *kb, double *fn, double *df) {

        for (uint64_t i=0; i<%(shape)d; ++i) {
            if (!mask[i]) {
                continue;
            }
            double x2 = x[i] * x[i];
            double x3 = x2 * x[i];
            double k12 = k1[i] * k2[i];
            double k12p = k1p[i] * k2p[i];
            double k123p = k12p * k3p[i];
            double c = 1.0 + st[i] / ks[i] + ft[i] / kf[i];
            double a = x3 + k1p[i] * x2 + k12p * x[i] + k123p;
            double a2 = a * a;
            double da = 3.0 * x2 + 2.0 * k1p[i] * x[i] + k12p;
            double b = x2 + k1[i] * x[i] + k12;
            double b2 = b * b;
            double db = 2.0 * x[i] + k1[i];


            fn[i] = k1[i] * x[i] * dic[i] / b
                    + 2.0 * dic[i] * k12 / b
                    + bt[i] / (1.0 + x[i] / kb[i])
                    + kw[i] / x[i]
                    + pt[i] * k12p * x[i] / a
                    + 2.0 * pt[i] * k123p / a
                    + sit[i] / (1.0 + x[i] / ksi[i])
                    - x[i] / c
                    - st[i] / (1.0 + ks[i] / (x[i] / c))
                    - ft[i] / (1.0 + kf[i] / (x[i] / c))
                    - pt[i] * x3 / a - ta[i];

            double t1 = (1.0 + x[i] / kb[i]);
            double t2 = (1.0 + x[i] / ksi[i]);
            double t3 = (1.0 + ks[i] / (x[i] / c));
            double t4 = (1.0 + kf[i] / (x[i] / c));

            df[i] = (k1[i] * dic[i] * b - k1[i] * x[i] * dic[i] * db) / b2
                    - 2.0 * dic[i] * k12 * db / b2
                    - bt[i] / kb[i] / (t1 * t1)
                    - kw[i] / x2
                    + (pt[i] * k12p * (a - x[i] * da)) / a2
                    - 2.0 * pt[i] * k123p * da / a2
                    - sit[i] / ksi[i] / (t2 * t2)
                    - 1.0 / c
                    - st[i] / (t3 * t3) * ks[i] * c / x2
                    - ft[i] / (t4 * t4) * kf[i] * c / x2
                    - pt[i] * x2 * (3.0 * a - x[i] * da) / a2;


        }
    }

    """ % {'shape': x.size}

    np.user_kernel.execute(kernel, arg_behaving + [fn, df])

    return fn, df


@veros_method
def ta_iter_SWS(*args):
    """
    Function to be optimized for free protons, calls Bohrium version if required


    Parameters
    ----------
    mask
        Mask indicating which cells to include / exclude from calculation
    x
        Free protons
    k1
    k2
    k1p
    k2p
    k3p
    st
    ks
    kf
    ft
    dic
    ta
    sit
    ksi
    pt
    bt
    kw
    kb

    Returns
    -------
    A function value and derivative in a tuple (f, df)
    """
    vs = args[0]
    mask = args[1]

    if rs.backend == 'bohrium':
        return ta_iter_SWS_bohrium(vs, mask, *args[2:])
    else:
        return ta_iter_SWS_numpy(vs, *args[2:])


@veros_method
def drtsafe_boundary_update(vs, mask, x_low, x_high, df, f, f_low, f_high, drtsafe_val):
    """
    For drtsafe: Moves the search boundary based on current result.

    Parameters
    ----------
    mask
        Mask for where to perform / skip calculations

    x_low
        Lower boundary value

    x_high
        Upper boundary value

    df
        Function derivative

    f
        Function value

    f_low
        Function value at lower bound

    f_high
        Function value at upper bound

    drtsafe_val
        Values for which f was calculated

    Returns
    -------
    None, but x_low, f_low, x_high, and f_high are updated to reflect new boundaries

    Note
    ----
    With the Bohrium backend, this function uses a user kernel to handle masked
    arrays. This is siginificantly faster than syncing to numpy or working on
    already completed cells.
    """

    if rs.backend == 'bohrium':
        mask_input = np.user_kernel.make_behaving(mask, dtype=np.bool)
        x_low_input = np.user_kernel.make_behaving(x_low, dtype=vs.default_float_type)
        x_high_input = np.user_kernel.make_behaving(x_high, dtype=vs.default_float_type)
        df_input = np.user_kernel.make_behaving(df, dtype=vs.default_float_type)
        f_input = np.user_kernel.make_behaving(f, dtype=vs.default_float_type)
        f_low_input = np.user_kernel.make_behaving(f_low, dtype=vs.default_float_type)
        f_high_input = np.user_kernel.make_behaving(f_high, dtype=vs.default_float_type)
        drtsafe_val_input = np.user_kernel.make_behaving(drtsafe_val, dtype=vs.default_float_type)

        kernel = """
        #include <stdint.h>
        #include <stdlib.h>
        #include <stdbool.h>
        void execute(bool *mask, double *x_low, double *x_high, double *df, double *f, double *f_low, double *f_high, double *drtsafe_val) {
            for (uint64_t i=0; i<%(shape0)d; ++i) {

                    if (!mask[i]) {
                        continue;
                    }
                    if (f[i] < 0.0) {
                        x_low[i] = drtsafe_val[i];
                        f_low[i] = f[i];
                    } else {
                        x_high[i] = drtsafe_val[i];
                        f_high[i] = f[i];
                    }
            }
        }
        """ % {'shape0': mask.shape[0] * mask.shape[1]}
        np.user_kernel.execute(kernel, [mask_input, x_low_input, x_high_input, df_input, f_input, f_low_input, f_high_input, drtsafe_val_input])

        # copy result back into variables
        x_low[...] = x_low_input[...]
        f_low[...] = f_low_input[...]
        x_high[...] = x_high_input[...]
        f_high[...] = f_high_input[...]

    else:
        x_low[mask], x_high[mask], f_low[mask], f_high[mask] = np.where(f[mask] < 0,
                (drtsafe_val[mask], x_high[mask], f[mask], f_high[mask]),
                (x_low[mask], drtsafe_val[mask], f_low[mask], f[mask]))


@veros_method
def drtsafe_step(vs, mask, drtsafe_val, x_high, df, f, x_low, dx, dx_old, accuracy):
    """
    Update step size

    If step size would take us out of bounds, set step size to half the interval length, else f/df
    and update step accordingly

    Parameters
    ----------
    mask
        Which cells should be calculated / skipped

    drtsafe_val
        Function input values

    x_high
        Upper bound of input values

    df
        Derivative in f

    f
        Function value

    x_low
        Lower bound of input values

    dx
        Change in function input value

    dx_old
        Change in function input value from previous step

    accuracy
        If calculated dx falls below this value, the mask for that cell will be updated

    Returns
    -------
    None, but mask, dtrsafe_val, dx, and dx_old will be updated


    Note
    ----
    With the Bohrium backend, this function uses a user kernel to skip completed cells.
    This is siginificantly faster than syncing to numpy or working on already completed cells.
    """

    if rs.backend == 'bohrium':
        mask_input = np.user_kernel.make_behaving(mask, dtype=np.int32)
        drtsafe_val_input = np.user_kernel.make_behaving(drtsafe_val, dtype=vs.default_float_type)
        x_high_input = np.user_kernel.make_behaving(x_high, dtype=vs.default_float_type)
        df_input = np.user_kernel.make_behaving(df, dtype=vs.default_float_type)
        f_input = np.user_kernel.make_behaving(f, dtype=vs.default_float_type)
        x_low_input = np.user_kernel.make_behaving(x_low, dtype=vs.default_float_type)
        dx_input = np.user_kernel.make_behaving(dx, dtype=vs.default_float_type)
        dx_old_input = np.user_kernel.make_behaving(dx_old, dtype=vs.default_float_type)

        kernel = """
        #include <stdint.h>
        #include <stdlib.h>
        #include <stdbool.h>
        #include <math.h>
        void execute(int *mask, double *drtsafe_val, double *x_high, double *df, double *f, double *x_low, double *dx, double *dx_old) {
            for (uint64_t i=0; i<%(shape0)d; ++i) {
                if (!mask[i]) {
                    continue;
                }

                double step1 = (drtsafe_val[i] - x_high[i]) * df[i] - f[i];
                double step2 = (drtsafe_val[i] - x_low[i]) * df[i] - f[i];
                bool step_inside = (step1 * step2) >= 0;
                bool tmp_mask = step_inside || (abs(2.0 * f[i]) > abs(dx_old[i] * df[i]));

                dx_old[i] = dx[i];
                if (tmp_mask) {
                    dx[i] = 0.5 * (x_high[i] - x_low[i]);
                    drtsafe_val[i] = x_low[i] + dx[i];
                } else {
                    dx[i] = f[i] / df[i];
                    drtsafe_val[i] = drtsafe_val[i] - dx[i];
                }
                mask[i] = fabs(dx[i]) > %(accuracy)E;
            }
        }
        """ % {'shape0': mask.size, 'accuracy': accuracy}
        np.user_kernel.execute(kernel, [mask_input, drtsafe_val_input, x_high_input, df_input,
                                        f_input, x_low_input, dx_input, dx_old_input])

        # Copy results back out
        mask[...] = mask_input[...]
        drtsafe_val[...] = drtsafe_val_input[...]
        dx[...] = dx_input[...]
        dx_old[...] = dx_old_input[...]

    else:
        tmp_mask = ((drtsafe_val[mask] - x_high[mask]) * df[mask] - f[mask]) * (
                    (drtsafe_val[mask] - x_low[mask]) * df[mask] - f[mask]) >= 0
        tmp_mask = np.logical_or(tmp_mask, (np.abs(2.0 * f[mask]) > np.abs(dx_old[mask] * df[mask])))

        dx_old[:] = dx.copy()
        dx[mask] = np.where(tmp_mask, 0.5 * (x_high[mask] - x_low[mask]), f[mask] / df[mask])
        drtsafe_val[mask] = np.where(tmp_mask, x_low[mask] + dx[mask], drtsafe_val[mask] - dx[mask])

        mask[mask] = np.abs(dx[mask]) > accuracy


@veros_method
def drtsafe_masked(vs, mask, function, guess_low, guess_high, args=None,
                   accuracy=1e-10, max_iterations=100):
    """
    Masked version of drtsafe

    Parameters
    ----------
    mask
        A mask indicating which cells may be skipped from calculations

    function
        Function to be optimzed. It must return a tuple consiting of function value and derivative

    guess_low
        Lower bound of initial guess bouding interval

    guess_high
        Upper bound of initial guess bounding interval

    args
        Additional fixed arguments passed to the function

    accuracy
        Maximum step size required as stop condition

    max_iterations
        Maximum number of optimization steps

    Returns
    -------
    Parameter values minimizing the function value
    """

    # Initial guess and step size
    drtsafe_val = 0.5 * (guess_low + guess_high)
    dx = np.abs(guess_high - guess_low)
    dx_old = dx.copy()

    # Function value at boundaries and guess
    f_low, _ = function(vs, mask, guess_low, *args)
    f_high, _ = function(vs, mask, guess_high, *args)
    f, df = function(vs, mask, drtsafe_val, *args)

    # Ensure low values have negative function value
    # and high has positive
    x_low, x_high, f_low, f_high = np.where(f_low < 0,
                                            (guess_low, guess_high, f_low, f_high),
                                            (guess_high, guess_low, f_high, f_low))

    for _ in range(max_iterations):

        # update step size, step and mask
        drtsafe_step(vs, mask, drtsafe_val, x_high, df, f, x_low, dx, dx_old, accuracy)

        if not mask.any():  # complete
            break

        # Update function for next step
        f, df = function(vs, mask, drtsafe_val, *args)

        # Update search boundaries
        drtsafe_boundary_update(vs, mask, x_low, x_high, df, f, f_low, f_high, drtsafe_val)

    return drtsafe_val


@veros_method
def drtsafe(vs, function, guess_low, guess_high, args=None, accuracy=1e-10, max_iterations=100):
    """ Root finding method with bounding box

    Follows Newton-Raphson step unless unsufficient or the step with take the solution out of bounds.
    Otherwise does birfurcating step.

    Parameters
    ----------
    function
        Function to be optimzed. It must return a tuple consiting of function value and derivative.

    guess_low
        Lower bound of initial guess bouding interval

    guess_high
        Upper bound of initial guess bounding interval

    args
        Additional fixed arguments passed to the function

    accuracy
        Maximum step size required as stop condition

    max_iterations
        Maximum number of optimization steps

    Returns
    -------
    Parameter values minimizing the function value
    """
    # Initial guess and step size
    drtsafe_val = 0.5 * (guess_low + guess_high)
    dx = np.abs(guess_high - guess_low)
    dx_old = dx.copy()

    # Function value at boundaries and guess
    f_low, _ = function(vs, guess_low, *args)
    f_high, _ = function(vs, guess_high, *args)
    f, df = function(vs, drtsafe_val, *args)

    # Ensure low values have negative function value
    # and high has positive
    x_low, x_high, f_low, f_high = np.where(f_low < 0,
                                            (guess_low, guess_high, f_low, f_high),
                                            (guess_high, guess_low, f_high, f_low))

    for _ in range(max_iterations):

        # update step size, step and mask
        step_mask = ((drtsafe_val - x_high) * df - f) * (
                    (drtsafe_val - x_low) * df - f) >= 0
        step_mask = np.logical_or(step_mask, (np.abs(2.0 * f) > np.abs(dx_old * df)))

        dx_old[:] = dx.copy()
        dx = np.where(step_mask, 0.5 * (x_high - x_low), f / df)

        if not (np.abs(dx) > accuracy).any():
            break

        # NOTE this was above the accuracy check
        drtsafe_val = np.where(step_mask, x_low + dx, drtsafe_val - dx)

        # Update function for next step
        f, df = function(vs, drtsafe_val, *args)

        # Update search boundaries
        x_low, x_high, f_low, f_high = np.where(f < 0,
                                                (drtsafe_val, x_high, f, f_high),
                                                (x_low, drtsafe_val, f_low, f))

    return drtsafe_val



from collections import OrderedDict

from veros.variables import Variable, T_GRID, T_HOR, YT, TIMESTEPS


MAIN_VARIABLES = OrderedDict([])

CONDITIONAL_VARIABLES = OrderedDict([
    ('enable_npzd', OrderedDict([
        ('bottom_mask', Variable(
            'Bottom mask', T_GRID, '', 'Bottom mask', dtype='int8'
        )),
        ('phytoplankton', Variable(
            'Phytoplankton concentration', T_GRID + TIMESTEPS, 'mmol/m^3?',
            'Concentration of phytoplankton in grid box',
            output=True,
            write_to_restart=True
        )),
        ('zooplankton', Variable(
            'Zooplankton concentration', T_GRID + TIMESTEPS, 'mmol/m^3?',
            'Concentration of zooplankton in grid box',
            output=True,
            write_to_restart=True
        )),
        ('detritus', Variable(
            'Detritus concentration', T_GRID + TIMESTEPS, 'mmol/m^3?',
            'Concentration of detritus in grid box',
            output=True,
            write_to_restart=True
        )),
        ('po4', Variable(
            'Phosphate concentration', T_GRID + TIMESTEPS, 'mmol/m^3?',
            'Concentration of phosphate in grid box',
            output=True,
            write_to_restart=True
        )),
        ('swr', Variable(
            'Shortwave radiation', T_HOR, 'W/m^3?',
            'Incomming solar radiation at sea level')),
        ('rctheta', Variable(
            'Effective vertical coordinate for incoming solar radiation', YT, '1',
            'Effective vertical coordinate for incoming solar radiation')),
        ('dayfrac', Variable(
            'Fraction of day with sunlight', YT, '1',
            'Fraction of day with sunlight')),
        ('excretion_total', Variable(
            'Total excretion from zooplankton', T_GRID, 'mmol/m^3 / s',
            'Zooplankton grazing causes excretion. This stores the total excreted amount for all consumed tracers')),
    ])),
    ('enable_carbon', OrderedDict([
        ('dic', Variable(
            'Dissolved Inorganic Carbon', T_GRID + TIMESTEPS, 'mmol/m^3',
            'Concentration of inorganic carbon ions and molecule',
            output=True,
            write_to_restart=True,
        )),
        ('alkalinity', Variable(
            'Alkalinity', T_GRID + TIMESTEPS, 'mmol/m^3',
            'Combined bases and acids',
            output=True,
            write_to_restart=True
        )),
        ('atmospheric_co2', Variable(
            'Atmospheric co2 concentration', T_HOR, 'ppmv',
            'Atmospheric co2 concentration')),
        ('cflux', Variable(
            'DIC Flux', T_HOR, 'mmol/m^3/s',
            'Flux of CO2 over the ocean-atmosphere bounday',
            output=True)),
        ('wind_speed', Variable(
            'Debugging wind speed', T_HOR, 'm/s',
            'Just used for debugging. Please ignore',
            output=True)),
        ('hSWS', Variable('hSWS', T_HOR, '1',
                          '[H] in Sea water sample', output=True)),
        ('pCO2', Variable('pCO2', T_HOR, '?ppmv/atm?',
                          'Partial CO2 pressure', output=True)),
        ('dpCO2', Variable('dpCO2', T_HOR, '?ppmv/atm?',
                           'Difference in ocean CO2 pressure and atmospheric', output=True)),
        ('co2star', Variable('co2star', T_HOR, '?ppmv?',
                             'Adjusted CO2 in ocean', output=True)),
        ('dco2star', Variable('dco2star', T_HOR, '?ppmv?',
                              'Adjusted CO2 difference', output=True)),
        ('rcak', Variable('Calcite redistribution share', T_GRID, '1',
                          'Calcite is redistributed after production by dissolution varying by depth')),
    ])),
])



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages
from codecs import open
import os

import versioneer


CLASSIFIERS = """
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: Implementation :: CPython
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

INSTALL_REQUIRES = [
    'veros>=0.2.1',
    'graphviz'
]

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='veros-bgc',
    license='MIT',
    author='Dion Häfner (NBI Copenhagen)',
    author_email='dion.haefner@nbi.ku.dk',
    keywords='oceanography python parallel numpy multi-core '
             'geophysics ocean-model bohrium mpi4py biogeochemistry',
    description='Biogeochemistry plugin for Veros, the versatile ocean simulator.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://veros-bgc.readthedocs.io',
    python_requires='>3.5.2',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    entry_points={
        'veros.setup_dirs': [
            'bgc = veros_bgc.setup'
        ]
    },
    classifiers=[c for c in CLASSIFIERS.split('\n') if c],
    zip_safe=False,
)



"""
Contains veros methods for handling bio- and geochemistry
(currently only simple bio)
"""
from collections import namedtuple

from veros import veros_method
from veros import time
from veros.core import diffusion, thermodynamics, utilities, isoneutral
from veros.variables import allocate


@veros_method
def biogeochemistry(vs):
    """Control function for integration of biogeochemistry

    Implements a rule based strategy. Any interaction between tracers should be
    described by registering a rule for the interaction.
    This routine ensures minimum tracer concentrations,
    calculates primary production, mortality, recyclability, vertical export for
    general tracers. Tracers should be registered to be used.
    """

    # Number of timesteps to do for bio tracers
    nbio = int(vs.dt_tracer // vs.dt_bio)

    # temporary tracer object to store differences
    for tracer, val in vs.npzd_tracers.items():
        vs.temporary_tracers[tracer][:, :, :] = val[:, :, :, vs.tau]

    # Flags enable us to only work on tracers with a minimum available concentration
    flags = {tracer: vs.maskT[...].astype(np.bool) for tracer in vs.temporary_tracers}

    # Pre rules: Changes that need to be applied before running npzd dynamics
    pre_rules = [(rule.function(vs, rule.source, rule.sink), rule.boundary)
                 for rule in vs.npzd_pre_rules]

    for rule, boundary in pre_rules:
        for key, value in rule.items():
            vs.temporary_tracers[key][boundary] += value

    # How much plankton is blocking light
    plankton_total = sum([plankton for plankton in vs.temporary_tracers.values()
                          if hasattr(plankton, 'light_attenuation')]) * vs.dzt

    # Integrated phytplankton - starting from top of layer going upwards
    # reverse cumulative sum because our top layer is the last.
    # Needs to be reversed again to reflect direction
    # phyto_integrated = np.empty_like(vs.temporary_tracers['phytoplankton'])
    phyto_integrated = np.empty_like(plankton_total)
    phyto_integrated[:, :, :-1] = plankton_total[:, :, 1:]
    phyto_integrated[:, :, -1] = 0.0

    # incomming shortwave radiation at top of layer
    swr = vs.swr[:, :, np.newaxis] * \
          np.exp(-vs.light_attenuation_phytoplankton
                 * np.cumsum(phyto_integrated[:, :, ::-1], axis=2)[:, :, ::-1])

    # Reduce incomming light where there is ice - as veros doesn't currently
    # have an ice model, we get temperatures below -1.8 and decreasing temperature forcing
    # as recommended by the 4deg model from the setup gallery
    icemask = np.logical_and(vs.temp[:, :, -1, vs.tau] * vs.maskT[:, :, -1] < -1.8, vs.forc_temp_surface < 0.0)
    swr[:, :] *= np.exp(-vs.light_attenuation_ice * icemask[:, :, np.newaxis])

    # declination and fraction of day with daylight
    # 0.72 is fraction of year at aphelion
    # 0.4 is scaling based on angle of rotation
    declin = np.sin((np.mod(vs.time * time.SECONDS_TO_X['years'], 1) - 0.72) * 2.0 * np.pi) * 0.4
    radian = 2 * np.pi / 360  # bohrium doesn't support np.radians, so this is faster
    rctheta = np.maximum(-1.5, np.minimum(1.5, vs.yt * radian - declin))

    # 1.33 is derived from Snells law for the air-sea barrier
    vs.rctheta[:] = vs.light_attenuation_water / np.sqrt(1.0 - (1.0 - np.cos(rctheta)**2.0) / 1.33**2)

    # fraction of day with photosynthetically active radiation with a minimum value
    dayfrac = np.minimum(1.0, -np.tan(radian * vs.yt) * np.tan(declin))
    vs.dayfrac[:] = np.maximum(1e-12, np.arccos(np.maximum(-1.0, dayfrac)) / np.pi)

    # light at top of grid box
    grid_light = swr * np.exp(vs.zw[np.newaxis, np.newaxis, :]
                              * vs.rctheta[np.newaxis, :, np.newaxis])

    # amount of PAR absorbed by water and plankton in each grid cell
    light_attenuation = vs.dzt * vs.light_attenuation_water +\
                        plankton_total * vs.light_attenuation_phytoplankton

    # common temperature factor determined according to b ** (cT)
    vs.bct = vs.bbio ** (vs.cbio * vs.temp[:, :, :, vs.tau])

    # light saturated growth and non-saturated growth
    jmax, avej = {}, {}
    for tracer in vs.temporary_tracers.values():

        # Calculate light limited vs unlimited growth
        if hasattr(tracer, 'potential_growth'):
            jmax[tracer.name], avej[tracer.name] = tracer.potential_growth(vs, grid_light, light_attenuation)

        # Methods for internal use may need an update
        if hasattr(tracer, 'update_internal'):
            tracer.update_internal(vs)

    # bio loop
    for _ in range(nbio):

        # Plankton is recycled, dying and growing
        # pre compute amounts for use in rules
        # for plankton in vs.plankton_types:
        for tracer in vs.temporary_tracers.values():

            # Nutrient limiting growth - if no limit, growth is determined by avej
            u = 1

            # limit maximum growth, usually by nutrient deficiency
            # and calculate primary production from that
            if hasattr(tracer, 'potential_growth'):
                # NOTE jmax and avej are NOT updated within bio loop
                for growth_limiting_function in vs.limiting_functions[tracer.name]:
                    u = np.minimum(u, growth_limiting_function(vs, vs.temporary_tracers))

                vs.net_primary_production[tracer.name] = flags[tracer.name] * flags['po4'] \
                    * np.minimum(avej[tracer.name], u * jmax[tracer.name]) * tracer

            # recycling methods - remineralization and fast recycling
            if hasattr(tracer, 'recycle'):
                vs.recycled[tracer.name] = flags[tracer.name] * tracer.recycle(vs)

            # Living tracers which can die
            if hasattr(tracer, 'mortality'):
                vs.mortality[tracer.name] = flags[tracer.name] * tracer.mortality(vs)

            # tracers which can graze on others
            if hasattr(tracer, 'grazing'):
                # TODO handle grazing by multiple zooplankton types
                # currently this only works with 1 type
                vs.grazing, vs.digestion, vs.excretion, vs.sloppy_feeding = tracer.grazing(vs, vs.temporary_tracers, flags)
                vs.excretion_total[...] = sum(vs.excretion.values())

            # Calculate concentration leaving cell and entering from above
            if hasattr(tracer, 'sinking_speed'):
                # Concentration of exported material is calculated as fraction
                # of total concentration which would have fallen through the bottom
                # of the cell (speed / vs.dzt * vs.dtbio)
                # vs.dtbio is accounted for later
                vs.npzd_export[tracer.name] = tracer.sinking_speed / vs.dzt * tracer * flags[tracer.name]

                # Import is export from above scaled by the ratio of cell heights
                vs.npzd_import[tracer.name] = np.empty_like(vs.npzd_export[tracer.name])
                vs.npzd_import[tracer.name][:, :, -1] = 0
                vs.npzd_import[tracer.name][:, :, :-1] = vs.npzd_export[tracer.name][:, :, 1:] * (vs.dzt[1:] / vs.dzt[:-1])

                # ensure we don't import in cells below bottom
                vs.npzd_import[tracer.name][...] *= vs.maskT

        # Gather all state updates
        npzd_updates = [(rule.function(vs, rule.source, rule.sink), rule.boundary)
                        for rule in vs.npzd_rules]

        # perform updates
        for update, boundary in npzd_updates:
            for key, value in update.items():
                vs.temporary_tracers[key][boundary] += value * vs.dt_bio

        # Import and export between layers
        # for tracer in vs.sinking_speeds:
        for tracer in vs.temporary_tracers.values():
            if hasattr(tracer, 'sinking_speed'):
                tracer[:, :, :] += (vs.npzd_import[tracer.name] - vs.npzd_export[tracer.name]) * vs.dt_bio

        # Prepare temporary tracers for next bio iteration
        for tracer, data in vs.temporary_tracers.items():
            flags[tracer][:, :, :] = np.logical_and(flags[tracer], (data > vs.trcmin))
            data[:, :, :] = utilities.where(vs, flags[tracer], data, vs.trcmin)

    # Post processesing or smoothing rules
    post_results = [(rule.function(vs, rule.source, rule.sink), rule.boundary)
                    for rule in vs.npzd_post_rules]
    post_modified = []  # we only want to reset values, which have acutally changed for performance

    for result, boundary in post_results:
        for key, value in result.items():
            vs.temporary_tracers[key][boundary] += value
            post_modified.append(key)

    # Reset before returning
    # using set for unique modifications is faster than resetting all tracers
    for tracer in set(post_modified):
        data = vs.temporary_tracers[tracer]
        flags[tracer][:, :, :] = np.logical_and(flags[tracer], (data > vs.trcmin))
        data[:, :, :] = utilities.where(vs, flags[tracer], data, vs.trcmin)

    # Only return the difference from the current time step. Will be added to timestep taup1
    return {tracer: vs.temporary_tracers[tracer] - vs.npzd_tracers[tracer][:, :, :, vs.tau]
            for tracer in vs.npzd_tracers}


def general_nutrient_limitation(nutrient, saturation_constant):
    """ Nutrient limitation form for all nutrients """
    return nutrient / (saturation_constant + nutrient)


@veros_method(inline=True)
def phosphate_limitation_phytoplankton(vs, tracers):
    """ Phytoplankton limit to growth by phosphate limitation """
    return general_nutrient_limitation(tracers['po4'], vs.saturation_constant_N * vs.redfield_ratio_PN)


@veros_method(inline=True)
def register_npzd_data(vs, tracer):
    """ Add tracer to the NPZD data set and create node in interaction graph

    Tracers added are available in the npzd dynamics and is automatically
    included in transport equations

    Parameters
    ----------
    tracer
        An instance of :obj:`veros.core.npzd_tracer.NPZD_tracer`
        to be included in biogeochemistry calculations
    """

    if tracer.name in vs.npzd_tracers.keys():
        raise ValueError('{name} has already been added to the NPZD data set'.format(name=tracer.name))

    vs.npzd_tracers[tracer.name] = tracer

    if tracer.transport:
        vs.npzd_transported_tracers.append(tracer.name)


@veros_method(inline=True)
def _get_boundary(vs, boundary_string):
    """ Return slice representing boundary

    Parameters
    ----------
    boundary_string
        Identifer for boundary. May take one of the following values:
        SURFACE:       [:, :, -1] only the top layer
        BOTTOM:        bottom_mask as set by veros
        else:          [:, :, :] everything
    """

    if boundary_string == 'SURFACE':
        return tuple([slice(None, None, None), slice(None, None, None), -1])

    if boundary_string == 'BOTTOM':
        return vs.bottom_mask

    return tuple([slice(None, None, None)] * 3)


@veros_method(inline=True)
def register_npzd_rule(vs, name, rule, label=None, boundary=None, group='PRIMARY'):
    """ Make rule available to the npzd dynamics

    The rule specifies an interaction between two tracers.
    It may additionally specify where in the grid it works as
    well as where in the execution order.

    Note
    ----
    Registering a rule is not sufficient for inclusion in dynamics.
    It must also be selected using select_npzd_rule.

    Parameters
    ----------
    name
        Unique identifier for the rule

    rule
        A list of rule names or tuple containing:
            function: function to be called
            source: what is being consumed
            destination: what is growing from consuming
    label
        A description for graph. See :obj:`veros.diagnostics.biogeochemistry`

    boundary
        'SURFACE', 'BOTTOM' or None, see _get_boundary

    group
        'PRIMARY' (default): Rule is evaluated in primary execution loop nbio times with timestepping
        'PRE': Rule is evaluated once before primary loop
        'POST': Rule is evaluated once after primary loop
    """

    Rule = namedtuple('Rule', ['function', 'source', 'sink', 'label', 'boundary', 'group'])

    if name in vs.npzd_available_rules:
        raise ValueError('Rule %s already exists, please verify the rule has not already been added and replace the chosen name' % name)

    if isinstance(rule, list):
        if label or boundary:
            raise ValueError('Cannot add labels or boundaries to rule groups')
        vs.npzd_available_rules[name] = rule

    else:
        label = label or '?'  # label is just for the interaction graph
        vs.npzd_available_rules[name] = Rule(function=rule[0], source=rule[1],
                                             sink=rule[2], label=label,
                                             boundary=_get_boundary(vs, boundary),
                                             group=group)

@veros_method(inline=True)
def register_npzd_common_source_rule(vs, name, rule, label=None, boundary=None, group='PRIMARY'):
    """ Register a rule to the model, which shares source term with other rules.

    This function creates stub rules for each individual term by evaluating the original rule and
    returning only the sink part of the original rule. Additionally there will be created a stub
    rule for the source term based on the first registered rule for the common source. Therefore a
    common source rule is best suited for rules, where the source terms has been preevaluated.
    The new registered rules are available with the name given by the parameter :obj:`name`
    postfixed by an underscore followed by the name of the tracer to be affected as specified in the
    rule. All created rules will be made available as a rule collection to be activated collectively
    with the name specified by the parameter :obj:`name`.
    Parameters are the same as register_npzd_rule

    Note
    ----
    Should only be used for rules, where the source term is exactly the same.

    Parameters
    ----------
    name
        Unique identifier for the rule

    rule
        A list of rule names or tuple containing:
            function: function to be called
            source: what is being consumed
            destination: what is growing from consuming
    label
        A description for graph. See :obj:`veros.diagnostics.biogeochemistry`

    boundary
        'SURFACE', 'BOTTOM' or None, see _get_boundary

    group
        'PRIMARY' (default): Rule is evaluated in primary execution loop nbio times with timestepping
        'PRE': Rule is evaluated once before primary loop
        'POST': Rule is evaluated once after primary loop
    """
    # The rule collection is made in setup_npzd based on the concents of common_source_rules

    if isinstance(rule, list):
        raise ValueError('register_npzd_common_source_rule does not accept list of rule names as rule parameter')

    # This should maybe be done elsewhere
    # Ensure that the common source rules dictionary exists
    # register_nzpd_rule handles not allowing multiple rules with the same name,
    # therefore also handling not having the same sink defined twice or the source also being a sink
    if not hasattr(vs, 'common_source_rules'):
        vs.common_source_rules = {}

    # If no rule for this name has been added yet, add the source term
    if name not in vs.common_source_rules.keys():
        vs.common_source_rules[name] = []

        # Add the source term
        # The new rule should consist of a new function, which is callable like the initial function of the rule, but only returns the source term
        source_rule = (veros_method(lambda veros, source, sink: {source: rule[0](veros, source, sink)[source]}, inline=True), rule[1], rule[2])
        register_npzd_rule(vs, f'{name}_{rule[1]}', source_rule, label=label + '(source)', boundary=boundary, group=group)

        vs.common_source_rules[name].append(source_rule)

    # Register new stub rule for the sink - source is assumed common so only added for the first
    sink_rule = (veros_method(lambda veros, source, sink: {sink: rule[0](veros, source, sink)[sink]}, inline=True), rule[1], rule[2])
    register_npzd_rule(vs, f'{name}_{rule[2]}', sink_rule, label=label, boundary=boundary, group=group)

    vs.common_source_rules[name].append(sink_rule)

    # Ensure that the registered rule works on the same boundary and same execution group as the source
    # The boundary is in principle not a strict requirement, but should in practice always be the same
    registered_source = vs.npzd_available_rules[f'{name}_{rule[1]}']
    registered_sink = vs.npzd_available_rules[f'{name}_{rule[2]}']

    if registered_source.boundary != registered_sink.boundary:
        raise ValueError(f'Common source rules must have the same boundary for both source and sink but {registered_source.name} did not match {registered_sink.name}. Expected {registered_source.boundary}, got {registed_sink.boundary}')
    if registered_source.group != registered_sink.group:
        raise ValueError(f'Common source rules must have the same group for both source and sink but {registered_source.name} did not match {registered_sink.name}. Expected {registered_source.group}, got {registed_sink.group}')

    # Make sure that the sources are actually listed as the same.
    # This check should not be necesarry if the user actually wants to register rules for a common source
    if registered_source.source != rule[1]:
        raise ValueError(f'Common source rules should have a common source, but {registered_source.name} did not match {registered_sink.name}. Expected {registered_source.source}, got {rule[1]}')


@veros_method(inline=True)
def select_npzd_rule(vs, name):
    """
    Select rule for the NPZD model

    Parameters
    ----------
    name
        Name of the rule to be selected
    """

    # activate rule by selecting from available rules
    rule = vs.npzd_available_rules[name]
    if name in vs.npzd_selected_rule_names:
        raise ValueError('Rules must have unique names, %s defined multiple times' % name)

    vs.npzd_selected_rule_names.append(name)

    # we may activate each rule in a list of rules
    if isinstance(rule, list):
        for r in rule:
            select_npzd_rule(vs, r)

    # or activate a single rule
    elif isinstance(rule, tuple):

        if rule.group == 'PRIMARY':
            vs.npzd_rules.append(rule)
        elif rule.group == 'PRE':
            vs.npzd_pre_rules.append(rule)
        elif rule.group == 'POST':
            vs.npzd_post_rules.append(rule)

    else:
        raise TypeError('Rule must be of type tuple or list')


@veros_method(inline=True)
def setup_basic_npzd_rules(vs):
    """
    Setup rules for basic NPZD model including phosphate, detritus, phytoplankton and zooplankton
    """
    from .npzd_rules import grazing, mortality, sloppy_feeding, recycling_to_po4, \
        primary_production, empty_rule, \
        bottom_remineralization_detritus_po4

    from .npzd_tracers import Recyclable_tracer, Phytoplankton, Zooplankton, NPZD_tracer

    # TODO - couldn't this be created elsewhere or can I use vs.kbot more efficiently?
    vs.bottom_mask[:, :, :] = np.arange(vs.nz)[np.newaxis, np.newaxis, :] == (vs.kbot - 1)[:, :, np.newaxis]

    zw = vs.zw - vs.dzt  # bottom of grid box using dzt because dzw is weird
    dtr_speed = (vs.wd0 + vs.mw * np.where(-zw < vs.mwz, -zw, vs.mwz)) \
        * vs.maskT

    detritus = Recyclable_tracer(vs.detritus, 'detritus',
                                 sinking_speed=dtr_speed,
                                 recycling_rate=vs.remineralization_rate_detritus)

    phytoplankton = Phytoplankton(vs.phytoplankton, 'phytoplankton',
                                  light_attenuation=vs.light_attenuation_phytoplankton,
                                  growth_parameter=vs.maximum_growth_rate_phyto,
                                  recycling_rate=vs.fast_recycling_rate_phytoplankton,
                                  mortality_rate=vs.specific_mortality_phytoplankton)

    zooplankton = Zooplankton(vs.zooplankton, 'zooplankton',
                              max_grazing=vs.maximum_grazing_rate,
                              grazing_saturation_constant=vs.saturation_constant_Z_grazing,
                              assimilation_efficiency=vs.assimilation_efficiency,
                              growth_efficiency=vs.zooplankton_growth_efficiency,
                              grazing_preferences=vs.zprefs,
                              mortality_rate=vs.quadric_mortality_zooplankton)

    po4 = NPZD_tracer(vs.po4, 'po4')

    # Add tracers to the model
    register_npzd_data(vs, detritus)
    register_npzd_data(vs, phytoplankton)
    register_npzd_data(vs, zooplankton)
    register_npzd_data(vs, po4)

    vs.limiting_functions['phytoplankton'] = [phosphate_limitation_phytoplankton]

    # Zooplankton preferences for grazing on keys
    # Values are scaled automatically at the end of this function
    vs.zprefs['phytoplankton'] = vs.zprefP
    vs.zprefs['zooplankton'] = vs.zprefZ
    vs.zprefs['detritus'] = vs.zprefDet

    # Register rules for interactions between active tracers
    register_npzd_rule(vs, 'npzd_basic_phytoplankton_grazing',
                       (grazing, 'phytoplankton', 'zooplankton'),
                       label='Grazing')
    register_npzd_rule(vs, 'npzd_basic_phytoplankton_mortality',
                       (mortality, 'phytoplankton', 'detritus'),
                       label='Mortality')
    register_npzd_rule(vs, 'npzd_basic_phytoplankton_fast_recycling',
                       (recycling_to_po4, 'phytoplankton', 'po4'),
                       label='Fast recycling')
    register_npzd_rule(vs, 'npzd_basic_zooplankton_grazing',
                       (empty_rule, 'zooplankton', 'zooplankton'),
                       label='Grazing')
    register_npzd_rule(vs, 'npzd_basic_zooplankton_mortality',
                       (mortality, 'zooplankton', 'detritus'),
                       label='Mortality')
    register_npzd_rule(vs, 'npzd_basic_zooplankton_sloppy_feeding',
                       (sloppy_feeding, 'zooplankton', 'detritus'),
                       label='Sloppy feeding')
    register_npzd_rule(vs, 'npzd_basic_detritus_grazing',
                       (grazing, 'detritus', 'zooplankton'),
                       label='Grazing')
    register_npzd_rule(vs, 'npzd_basic_detritus_remineralization',
                       (recycling_to_po4, 'detritus', 'po4'),
                       label='Remineralization')
    register_npzd_rule(vs, 'npzd_basic_phytoplankton_primary_production',
                       (primary_production, 'po4', 'phytoplankton'),
                       label='Primary production')

    register_npzd_rule(vs, 'npzd_basic_detritus_bottom_remineralization',
                       (bottom_remineralization_detritus_po4, 'detritus', 'po4'),
                       label='Bottom remineralization', boundary='BOTTOM')

    register_npzd_rule(vs, 'group_npzd_basic', [
        'npzd_basic_phytoplankton_grazing',
        'npzd_basic_phytoplankton_mortality',
        'npzd_basic_phytoplankton_fast_recycling',
        'npzd_basic_phytoplankton_primary_production',
        'npzd_basic_zooplankton_grazing',
        'npzd_basic_zooplankton_mortality',
        'npzd_basic_zooplankton_sloppy_feeding',
        'npzd_basic_detritus_remineralization',
        'npzd_basic_detritus_grazing',
        'npzd_basic_detritus_bottom_remineralization'
    ])

    register_npzd_rule(vs, 'empty_rule', (empty_rule, None, None))


@veros_method(inline=True)
def setup_carbon_npzd_rules(vs):
    """
    Rules for including a carbon cycle
    """
    # The actual action is on DIC, but the to variables overlap
    from .npzd_rules import co2_surface_flux, recycling_to_dic, \
        primary_production_from_DIC, excretion_dic, recycling_phyto_to_dic, \
        dic_alk_scale, calcite_production_phyto, calcite_production_phyto_alk, \
        post_redistribute_calcite, post_redistribute_calcite_alk, pre_reset_calcite, \
        bottom_remineralization_detritus_DIC

    from .npzd_tracers import NPZD_tracer

    zw = vs.zw - vs.dzt  # bottom of grid box using dzt because dzw is weird

    # redistribution fraction for calcite at level k
    vs.rcak[:, :, :-1] = (- np.exp(zw[:-1] / vs.dcaco3) + np.exp(zw[1:] / vs.dcaco3)) / vs.dzt[:-1]
    vs.rcak[:, :, -1] = - (np.exp(zw[-1] / vs.dcaco3) - 1.0) / vs.dzt[-1]

    # redistribution fraction at bottom
    rcab = np.empty_like(vs.dic[..., 0])
    rcab[:, : -1] = 1 / vs.dzt[-1]
    rcab[:, :, :-1] = np.exp(zw[:-1] / vs.dcaco3) / vs.dzt[1:]

    # merge bottom into level k and reset every cell outside ocean
    vs.rcak[vs.bottom_mask] = rcab[vs.bottom_mask]
    vs.rcak[...] *= vs.maskT

    # Need to track dissolved inorganic carbon, alkalinity
    dic = NPZD_tracer(vs.dic, 'DIC')
    alk = NPZD_tracer(vs.alkalinity, 'alkalinity')
    register_npzd_data(vs, dic)
    register_npzd_data(vs, alk)

    # Only for collection purposes - to be redistributed in post rules
    caco3 = NPZD_tracer(np.zeros_like(vs.dic), "caco3", transport=False)
    register_npzd_data(vs, caco3)

    # Atmosphere
    register_npzd_rule(vs, 'npzd_carbon_flux',
                       (co2_surface_flux, 'co2', 'DIC'),
                       boundary='SURFACE', group='PRE',
                       label='Atmosphere exchange')

    # Common rule set for nutrient
    register_npzd_rule(vs, 'npzd_carbon_recycling_detritus_dic',
                       (recycling_to_dic, 'detritus', 'DIC'),
                       label='Remineralization')
    register_npzd_rule(vs, 'npzd_carbon_primary_production_dic',
                       (primary_production_from_DIC, 'DIC', 'phytoplankton'),
                       label='Primary production')
    register_npzd_rule(vs, 'npzd_carbon_recycling_phyto_dic',
                       (recycling_phyto_to_dic, 'phytoplankton', 'DIC'),
                       label='Fast recycling')
    register_npzd_rule(vs, 'npzd_carbon_excretion_dic',
                       (excretion_dic, 'zooplankton', 'DIC'),
                       label='Excretion')
    register_npzd_rule(vs, 'npzd_carbon_calcite_production_dic',
                       (calcite_production_phyto, 'DIC', 'caco3'),
                       label='Production of calcite')
    register_npzd_rule(vs, 'npzd_carbon_calcite_production_alk',
                       (calcite_production_phyto_alk, 'alkalinity', 'caco3'),
                       label='Production of calcite')

    # BOTTOM
    register_npzd_rule(vs, 'npzd_carbon_detritus_bottom_remineralization',
                       (bottom_remineralization_detritus_DIC, 'detritus', 'DIC'),
                       label='Bottom remineralization', boundary='BOTTOM')

    # POST rules
    register_npzd_rule(vs, 'npzd_carbon_dic_alk',
                       (dic_alk_scale, 'DIC', 'alkalinity'),
                       group='POST',
                       label='Changes in DIC reflected in ALK')
    register_npzd_rule(vs, 'npzd_carbon_post_distribute_calcite_alk',
                       (post_redistribute_calcite_alk, 'caco3', 'alkalinity'),
                       label='dissolution', group='POST')
    register_npzd_rule(vs, 'npzd_carbon_post_distribute_calcite_dic',
                       (post_redistribute_calcite, 'caco3', 'DIC'),
                       label='dissolution', group='POST')

    # PRE
    register_npzd_rule(vs, 'pre_reset_calcite',
                       (pre_reset_calcite, 'caco3', 'caco3'),
                       label='reset', group='PRE')

    register_npzd_rule(vs, 'group_carbon_implicit_caco3', [
        'npzd_carbon_flux',
        'npzd_carbon_recycling_detritus_dic',
        'npzd_carbon_primary_production_dic',
        'npzd_carbon_recycling_phyto_dic',
        'npzd_carbon_excretion_dic',
        'npzd_carbon_dic_alk',
        'npzd_carbon_calcite_production_dic',
        'npzd_carbon_calcite_production_alk',
        'npzd_carbon_post_distribute_calcite_alk',
        'npzd_carbon_post_distribute_calcite_dic',
        'npzd_carbon_detritus_bottom_remineralization',
        'pre_reset_calcite',
    ])


@veros_method
def setupNPZD(vs):
    """Taking veros variables and packaging them up into iterables"""
    if not vs.enable_npzd:
        return

    setup_basic_npzd_rules(vs)

    # Add carbon to the model
    if vs.enable_carbon:
        setup_carbon_npzd_rules(vs)

    from .npzd_rules import excretion
    register_npzd_common_source_rule(vs, 'npzd_basic_zooplankton_excretion',
                                     (excretion, 'zooplankton', 'po4'),
                                     label='Excretion')

    register_npzd_common_source_rule(vs, 'npzd_basic_zooplankton_excretion',
                                     (excretion, 'zooplankton', 'DIC'),
                                     label='Excretion')

    # Turn common source rules into selectable rules
    # We should not have to make this check, it should just be defined
    if hasattr(vs, 'common_source_rules'):
        # All common source rules have been saved to a dictionary
        # where the key is the collection identifier and individual rules are
        # named after the convention {collection_name}_{tracer_name}
        for name, rules in vs.common_source_rules.items():
            collection = [name + "_" + rules[0][1]] + [name + "_" + rules[i][2] for i in range(1, len(rules))]
            register_npzd_rule(vs, name, collection)

    for rule in vs.npzd_selected_rules:
        select_npzd_rule(vs, rule)

    # Update Zooplankton preferences dynamically
    # Ideally this would be done in the Zooplankton class
    zprefsum = sum(vs.zprefs.values())
    for preference in vs.zprefs:
        vs.zprefs[preference] /= zprefsum

    # Keep derivatives of everything for advection
    for tracer, data in vs.npzd_tracers.items():
        vs.npzd_advection_derivatives[tracer] = np.zeros_like(data)

    # Temporary tracers are necessary to only return differences
    for tracer, data in vs.npzd_tracers.items():
        vs.temporary_tracers[tracer] = np.empty_like(data[..., 0])


@veros_method
def npzd(vs):
    r"""
    Main driving function for NPZD functionality

    Computes transport terms and biological activity separately

    \begin{equation}
        \dfrac{\partial C_i}{\partial t} = T + S
    \end{equation}
    """
    if not vs.enable_npzd:
        return

    # TODO: Refactor transportation code to be defined only once and also used by thermodynamics
    # TODO: Dissipation on W-grid if necessary

    npzd_changes = biogeochemistry(vs)

    """
    For vertical mixing
    """

    a_tri = allocate(vs, ('xt', 'yt', 'zt'), include_ghosts=False)
    b_tri = allocate(vs, ('xt', 'yt', 'zt'), include_ghosts=False)
    c_tri = allocate(vs, ('xt', 'yt', 'zt'), include_ghosts=False)
    d_tri = allocate(vs, ('xt', 'yt', 'zt'), include_ghosts=False)
    delta = allocate(vs, ('xt', 'yt', 'zt'), include_ghosts=False)

    ks = vs.kbot[2:-2, 2:-2] - 1
    delta[:, :, :-1] = vs.dt_tracer / vs.dzw[np.newaxis, np.newaxis, :-1]\
        * vs.kappaH[2:-2, 2:-2, :-1]
    delta[:, :, -1] = 0
    a_tri[:, :, 1:] = -delta[:, :, :-1] / vs.dzt[np.newaxis, np.newaxis, 1:]
    b_tri[:, :, 1:] = 1 + (delta[:, :, 1:] + delta[:, :, :-1]) / vs.dzt[np.newaxis, np.newaxis, 1:]
    b_tri_edge = 1 + delta / vs.dzt[np.newaxis, np.newaxis, :]
    c_tri[:, :, :-1] = -delta[:, :, :-1] / vs.dzt[np.newaxis, np.newaxis, :-1]

    for tracer in vs.npzd_transported_tracers:
        tracer_data = vs.npzd_tracers[tracer]

        """
        Advection of tracers
        """
        thermodynamics.advect_tracer(vs, tracer_data[:, :, :, vs.tau],
                                     vs.npzd_advection_derivatives[tracer][:, :, :, vs.tau])

        # Adam-Bashforth timestepping
        tracer_data[:, :, :, vs.taup1] = tracer_data[:, :, :, vs.tau] + vs.dt_tracer \
            * ((1.5 + vs.AB_eps) * vs.npzd_advection_derivatives[tracer][:, :, :, vs.tau]
               - (0.5 + vs.AB_eps) * vs.npzd_advection_derivatives[tracer][:, :, :, vs.taum1])\
            * vs.maskT

        """
        Diffusion of tracers
        """

        if vs.enable_hor_diffusion:
            horizontal_diffusion_change = np.zeros_like(tracer_data[:, :, :, 0])
            diffusion.horizontal_diffusion(vs, tracer_data[:, :, :, vs.tau],
                                           horizontal_diffusion_change)

            tracer_data[:, :, :, vs.taup1] += vs.dt_tracer * horizontal_diffusion_change

        if vs.enable_biharmonic_mixing:
            biharmonic_diffusion_change = np.empty_like(tracer_data[:, :, :, 0])
            diffusion.biharmonic(vs, tracer_data[:, :, :, vs.tau],
                                 np.sqrt(abs(vs.K_hbi)), biharmonic_diffusion_change)

            tracer_data[:, :, :, vs.taup1] += vs.dt_tracer * biharmonic_diffusion_change

        """
        Restoring zones
        """
        # TODO add restoring zones to general tracers

        """
        Isopycnal diffusion
        """
        if vs.enable_neutral_diffusion:
            dtracer_iso = np.zeros_like(tracer_data[..., 0])

            isoneutral.isoneutral_diffusion_tracer(vs, tracer_data, dtracer_iso,
                                                   iso=True, skew=False)

            if vs.enable_skew_diffusion:
                dtracer_skew = np.zeros_like(tracer_data[..., 0])
                isoneutral.isoneutral_diffusion_tracer(vs, tracer_data, dtracer_skew,
                                                       iso=False, skew=True)
        """
        Vertical mixing of tracers
        """
        d_tri[:, :, :] = tracer_data[2:-2, 2:-2, :, vs.taup1]
        # TODO: surface flux?
        # d_tri[:, :, -1] += surface_forcing
        sol, mask = utilities.solve_implicit(vs, ks, a_tri, b_tri, c_tri, d_tri, b_edge=b_tri_edge)

        tracer_data[2:-2, 2:-2, :, vs.taup1] = utilities.where(vs, mask, sol,
                                                               tracer_data[2:-2, 2:-2, :, vs.taup1])

    # update by biogeochemical changes
    for tracer, change in npzd_changes.items():
        vs.npzd_tracers[tracer][:, :, :, vs.taup1] += change

    # prepare next timestep with minimum tracer values
    for tracer in vs.npzd_tracers.values():
        tracer[:, :, :, vs.taup1] = np.maximum(tracer[:, :, :, vs.taup1], vs.trcmin * vs.maskT)

    for tracer in vs.npzd_tracers.values():
        utilities.enforce_boundaries(vs, tracer)
